from functools import partial

from elasticai.creator.tests.vhdl.vhdl_file_testcase import GeneratedVHDLCodeTest
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import (
    PrecomputedScalarTestBench,
)


class SigmoidTestBenchTest(GeneratedVHDLCodeTest):
    def test_compare_files(self) -> None:
        expected_code = """
        library ieee;
        use ieee.std_logic_1164.all;
        use ieee.numeric_std.all;
        use ieee.math_real.all;

        entity sigmoid_tb is
            generic (
                DATA_WIDTH : integer := 16;
                FRAC_WIDTH : integer := 0
            );
            port (
                clk : out std_logic
            );
        end entity sigmoid_tb;

        architecture rtl of sigmoid_tb is

            signal clk_period : time := 1 ns;
            signal test_input : signed(16-1 downto 0):=(others=>'0');
            signal test_output : signed(16-1 downto 0);

            component sigmoid is
                generic (
                    DATA_WIDTH : integer := 16;
                    FRAC_WIDTH : integer := 0
                );
                port (
                    x : in signed(DATA_WIDTH-1 downto 0);
                    y : out signed(DATA_WIDTH-1 downto 0)
                );
            end component sigmoid;

        begin

            clock_process: process
            begin
                clk <= '0';
                wait for clk_period/2;
                clk <= '1';
                wait for clk_period/2;
            end process clock_process;

            uut: sigmoid
            port map (
                x => test_input,
                y => test_output
            );

            test_process: process
            begin
                report "======Simulation Start======" severity Note;

                test_input <= to_signed(-1281,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=0 report "The test case -1281 fail" severity failure;

                test_input <= to_signed(-1000,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=4 report "The test case -1000 fail" severity failure;

                test_input <= to_signed(-500,16);
                wait for 1*clk_period;
                report "The value of 'test_output' is " & integer'image(to_integer(unsigned(test_output)));
                assert test_output=28 report "The test case -500 fail" severity failure;


                report "======Simulation Success======" severity Note;
                report "Please check the output message." severity Note;

                wait;

            end process test_process;

        end architecture rtl;
        """
        to_fp = partial(float_values_to_fixed_point, total_bits=16, frac_bits=0)
        sigmoid = PrecomputedScalarTestBench(
            component_name="sigmoid",
            x_list_for_testing=to_fp([-1281, -1000, -500]),
            y_list_for_testing=to_fp([0, 4, 28]),
        )
        sigmoid_code = sigmoid()
        sigmoid_code_str = ""
        for line in sigmoid_code:
            sigmoid_code_str += line + "\n"
        self.check_generated_code(expected_code, sigmoid_code_str)



import types
from abc import abstractmethod
from itertools import product
from typing import Any, Callable, Optional, Protocol

import torch
from torch import Tensor
from torch.nn import BatchNorm1d, Conv1d, Module, Parameter
from torch.nn.utils.parametrize import register_parametrization

from elasticai.creator.precomputation.input_domains import (
    create_codomain_for_1d_conv,
    create_codomain_for_depthwise_1d_conv,
)
from elasticai.creator.precomputation.precomputation import precomputable
from elasticai.creator.qat.functional import binarize as BinarizeFn
from elasticai.creator.tags_utils import TaggedModule

"""Implementation of quantizers and quantized variants of pytorch layers"""


class Quantize(Protocol):
    """
    Protocol to be adhered to by quantizations with properties to simplify translation to other levels of abstraction.
    Codomain is used for precalculation of convolutions by obtaining the codomain of the quantization prior to the convolution.
    Threshold is used on MVTUs.
    """

    def __call__(self, *args: Any, **kwargs: Any) -> Tensor:
        ...

    @property
    @abstractmethod
    def codomain(self) -> list[float]:
        ...

    @property
    @abstractmethod
    def thresholds(self) -> list[float]:
        ...


class Binarize(torch.nn.Module):
    """
    Implementation of binarizer with possible bits [-1,1]"""

    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(x):
        return BinarizeFn.apply(x)

    @property
    def codomain(self):
        return [-1, 1]

    @property
    def thresholds(self):
        return torch.Tensor([0.0])

    @staticmethod
    def right_inverse(x: Tensor) -> Tensor:
        return x


class Ternarize(torch.nn.Module):
    """
    Implementation of ternarizer with possible bits [-1,0,1]
    Args:
     zero_window_width: all numbers x with |x| <= |zero_window_width| will be rounded to zero, *default* value is
     0.5 which corresponds to natural rounding
    """

    def __init__(
        self,
        zero_window_width=0.5,
        trainable=False,
    ):
        super().__init__()

        factor = torch.tensor(0.5 / zero_window_width)

        # naming only possible per axis and is only a prototype feature --> decline
        # requires_grad flag to false makes parameters non-trainable
        self.widening_factor = torch.nn.Parameter(factor, requires_grad=trainable)

    def forward(self, x):
        clipped = torch.clamp(self.widening_factor * x, min=-1, max=1)
        rounded = torch.round(clipped)
        with torch.no_grad():
            no_gradient = rounded - clipped
        return clipped + no_gradient

    @property
    def codomain(self):
        return [-1, 0, 1]

    @staticmethod
    def right_inverse(x: Tensor) -> Tensor:
        return x

    @property
    def thresholds(self):
        return torch.Tensor([0.5 / self.widening_factor, -0.5 / self.widening_factor])


class ResidualQuantization(torch.nn.Module):
    """
    Implementation of residual quantization similar to Rebnet: https://arxiv.org/pdf/1711.01243.pdf
    But instead of summing the bits, they are concatenated, multiplying the second dimension by n bits.
    Here the encoded form is used to facilitate precalculation.
    Args:
     num_bits: number of residual quantization levels or bits
    """

    def __init__(self, num_bits=2):
        super().__init__()
        self.num_bits = num_bits
        self.weight = torch.nn.Parameter(
            torch.Tensor([0.5**i for i in range(1, num_bits)])
        )

    def forward(self, inputs):
        bin = BinarizeFn.apply
        out = [inputs]
        for factor in self.weight:
            out.append(out[-1] - torch.abs(factor) * bin(out[-1]))
        return bin(torch.cat(out, dim=1))

    @property
    def codomain(self):
        return list(product([-1, 1], repeat=self.num_bits))

    @property
    def thresholds(self):
        threshold = [torch.Tensor([0.0])]
        for factor in self.weight:
            threshold.append(threshold[-1] + torch.abs(factor))

        return torch.flatten(torch.stack(threshold, 0))


class QuantizeTwoBit(torch.nn.Module):
    """
    Implementation of 2 bit residual quantization, the first bit is binarized the second uses  a factor similar to residual quantization
    Args:
      factors: the initial factor
    """

    def __init__(self, factors=0.5):
        super().__init__()
        self.factors = torch.nn.Parameter(torch.Tensor([factors]), requires_grad=True)

    def forward(self, input):
        binarize = BinarizeFn.apply
        first_half = binarize(input)
        second_half = input - self.factors * binarize(input)
        return binarize(torch.cat((first_half, second_half), dim=1))

    @property
    def codomain(self):
        return [[x, y] for x in [-1, 1] for y in [-1, 1]]

    @property
    def thresholds(self):
        return torch.cat((torch.Tensor([0.0]), self.factors), 0)


def _init_quantizable_convolution(self, quantizer, bias, constraints):
    if isinstance(quantizer, Module):
        register_parametrization(self, "weight", quantizer)
        if bias:
            register_parametrization(self, "bias", quantizer)
    else:
        raise TypeError(f"Quantizer {quantizer} is not an instance of Module.")
    self.constraints = constraints if constraints else None

    def apply_constraint(self):
        if self.constraints:
            [constraint(self) for constraint in self.constraints]

    self.apply_constraint = types.MethodType(apply_constraint, self)


class QConv1d(torch.nn.Conv1d):
    """
    Implementation of quantized Conv1d layer,all parameters are equivalent to the base pytorch class except for the quantizer, and constraints.
    Args:
      quantizer: An instance of a quantizer for weight and bias , currently only 1 can be used for both
      constraints: A list of instances of constraints, applied with the apply constraint_call
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int],
        quantizer: Module,
        stride: tuple[int] = (1,),
        padding: int = 0,
        dilation: tuple[int] = (1,),
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        constraints: list = None,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )

        _init_quantizable_convolution(
            self, quantizer=quantizer, bias=bias, constraints=constraints
        )


class QConv2d(torch.nn.Conv2d):
    """
    Implementation of quantized Conv2d layer, all parameters are equivalent to the base pytorch class except for the quantizer and consraints.
    Args:
     quantizer: An instance of a quantizer for weight and bias , currently only 1 can be used for both
    constraints: A list of instances of constraints, applied with the apply constraint_call
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        quantizer,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        constraints: list = None,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )
        _init_quantizable_convolution(
            self, quantizer=quantizer, bias=bias, constraints=constraints
        )


class ChannelShuffle(torch.nn.Module):
    def __init__(self, groups):
        super(ChannelShuffle, self).__init__()
        self.groups = groups

    def forward(self, x):
        num_channels = x.data.size()[1]
        original_shape = x.data.size()[2:]
        x = x.view(
            x.data.size()[0], self.groups, num_channels // self.groups, *original_shape
        )
        x = x.transpose(1, 2).contiguous()
        return x.view(x.data.size()[0], num_channels, *original_shape)


class QLinear(torch.nn.Linear):
    """
    Implementation of quantized Linear layer,all parameters are equivalent to the base pytorch class except for the quantizer and constraints.
    Args:
     quantizer: An instance of a quantizer for weight and bias, currently only 1 can be used for both
     constraints: A list of instances of constraints, applied with the apply constraint_call
    """

    def __init__(
        self, in_features, out_features, quantizer, bias=True, constraints: list = None
    ):
        super().__init__(in_features=in_features, out_features=out_features, bias=bias)
        _init_quantizable_convolution(
            self, quantizer=quantizer, bias=bias, constraints=constraints
        )


class QLSTMCell(torch.nn.LSTMCell):
    """
    Implementation of quantized LSTM cell,all parameters are equivalent to the base pytorch class except for the quantizers.
    Args:
        input_gate_activation:  The input gate activation, expects a quantizer instance, if None will default to sigmoid
        forget_gate_activation: The forget gate activation, expects a quantizer instance, if None will default to sigmoid
        cell_gate_activation: The cell gate activation, expects a quantizer instance, if None will default to tanh
        output_gate_activation: The output gate activation, expects a quantizer instance, if None will default to sigmoid
        new_cell_gate_activation: The new cell gate activation, expects a quantizer instance, if None will default to tanh
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        state_quantizer: Quantize,
        weight_quantizer: Quantize,
        bias: bool = True,
        input_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        forget_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        cell_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.tanh,
        output_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        new_cell_state_activation: Callable[[torch.Tensor], torch.Tensor] = torch.tanh,
    ):
        super().__init__(input_size, hidden_size, bias)

        self.state_quantizer = state_quantizer
        self.weight_quantizer = weight_quantizer
        self.input_gate_activation = input_gate_activation
        self.forget_gate_activation = forget_gate_activation
        self.cell_gate_activation = cell_gate_activation
        self.output_gate_activation = output_gate_activation
        self.new_cell_state_activation = new_cell_state_activation

    @staticmethod
    def __identity(input_tensor: torch.Tensor) -> torch.Tensor:
        return input_tensor

    def forward(
        self, x: torch.Tensor, hx: Optional[tuple[torch.Tensor, torch.Tensor]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Implementation based on
        # https://github.com/pytorch/pytorch/blob/e9ef087d2d12051341db485c8ac64ea64649823d/benchmarks/fastrnns/cells.py#L25
        if hx is None:
            zeros = torch.zeros(
                x.size(0), self.hidden_size, dtype=x.dtype, device=x.device
            )
            hx = (zeros, zeros)

        h_0, c_0 = self.state_quantizer(hx[0]), self.state_quantizer(hx[1])
        weight_ih, weight_hh = self.weight_quantizer(
            self.weight_ih
        ), self.weight_quantizer(self.weight_hh)

        if self.bias:
            bias_ih, bias_hh = self.weight_quantizer(
                self.bias_ih
            ), self.weight_quantizer(self.bias_hh)
            gates = (
                torch.mm(x, weight_ih.t())
                + bias_ih
                + torch.mm(h_0, weight_hh.t())
                + bias_hh
            )
        else:
            gates = torch.mm(x, weight_ih.t()) + torch.mm(h_0, weight_hh.t())

        in_gate, forget_gate, cell_gate, out_gate = gates.chunk(4, dim=1)

        i = self.input_gate_activation(in_gate)
        f = self.forget_gate_activation(forget_gate)
        g = self.cell_gate_activation(cell_gate)
        o = self.output_gate_activation(out_gate)

        c_1 = f * c_0 + i * g
        h_1 = o * self.new_cell_state_activation(c_1)

        return h_1, c_1


class QLSTM(torch.nn.Module):
    """
    Implementation of quantized LSTM
    Args:
     lstm_cell : an instance of a lstm_cell
     batch_first: same as pytorch. If True, then the input and output tensors are provided as (batch, seq, feature) instead of (seq, batch, feature)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        state_quantizer: Quantize,
        weight_quantizer: Quantize,
        bias: bool = True,
        input_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        forget_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        cell_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.tanh,
        output_gate_activation: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
        new_cell_state_activation: Callable[[torch.Tensor], torch.Tensor] = torch.tanh,
        batch_first=False,
    ):
        super().__init__()
        self.cell = QLSTMCell(
            hidden_size=hidden_size,
            state_quantizer=state_quantizer,
            weight_quantizer=weight_quantizer,
            bias=bias,
            input_size=input_size,
            input_gate_activation=input_gate_activation,
            forget_gate_activation=forget_gate_activation,
            cell_gate_activation=cell_gate_activation,
            output_gate_activation=output_gate_activation,
            new_cell_state_activation=new_cell_state_activation,
        )
        self.batch_first = batch_first

    def forward(
        self, input: torch.Tensor, state: tuple[torch.Tensor, torch.Tensor] = None
    ) -> tuple[torch.Tensor, Optional[tuple[torch.Tensor, torch.Tensor]]]:
        if self.batch_first:
            input = torch.stack(torch.unbind(input), dim=1)

        # Implementation based on
        # https://github.com/pytorch/pytorch/blob/bb7fd1fcfbd2507272fd9b3f2610ef02bfba5692/benchmarks/fastrnns/custom_lstms.py#L184
        inputs = torch.unbind(input, dim=0)
        outputs = []
        for i in range(len(inputs)):
            output, cell_state = self.cell(inputs[i], state)
            state = output, cell_state
            outputs += [output]
        stack_dim = 1 if self.batch_first else 0
        result = torch.stack(outputs, dim=stack_dim)
        return result, state


class BatchNormedActivatedConv1d(torch.nn.Module):
    """Applies a convolution followed by a batchnorm and an activation function.
    The BatchNorm is not performing an affine translation, instead we incorporate
    a trainable scaling factor that is applied to each channel before the application
    of the activation function.
    """

    def __init__(
        self,
        activation: Callable[[], Quantize],
        kernel_size,
        in_channels,
        out_channels,
        groups,
        bias,
        channel_multiplexing_factor,
    ):
        super().__init__()
        self.conv = Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
            padding=0,
            bias=bias,
        )
        self.bn = BatchNorm1d(num_features=out_channels, affine=False)
        self.scaling_factors = Parameter(torch.ones((out_channels, 1)))
        self.quantize = activation()
        self.channel_multiplexing_factor = channel_multiplexing_factor

    @property
    def codomain_elements(self) -> list[float]:
        return self.quantize.codomain

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.scaling_factors * x
        x = self.quantize(x)
        return x


def define_batch_normed_convolution_1d(activation, channel_multiplexing_factor):
    """Allows to define a new `BatchNormedActivationConvolution` that uses `activation` for its activation function.
    The `channel_multiplexing_factor` will be used by calling modules to determine if the activation function will
    change the number of channels, like the `ResidualBinarization` does.
    This way we can determine the input shape for following layers."""

    def wrapper(cls):
        class Wrapped(BatchNormedActivatedConv1d):
            def __init__(self, kernel_size, in_channels, out_channels, groups, bias):
                super().__init__(
                    activation=activation,
                    kernel_size=kernel_size,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    groups=groups,
                    bias=bias,
                    channel_multiplexing_factor=channel_multiplexing_factor,
                )

        return Wrapped

    return wrapper


@define_batch_normed_convolution_1d(Ternarize, 1)
class TernaryConvolution1d:
    pass


@define_batch_normed_convolution_1d(QuantizeTwoBit, 2)
class MultilevelResidualBinarizationConv1d:
    pass


@define_batch_normed_convolution_1d(Binarize, 1)
class BinaryActivatedConv1d:
    pass


class SplitConvolutionBase(torch.nn.Module):
    def __init__(
        self,
        convolution: TaggedModule,
        kernel_size,
        in_channels,
        out_channels,
        codomain_elements: list[float],
    ):
        """Use this as a base class if you want to incorporate your own convolution implementation (up to the quantizing activation function)
        into a SplitConvolution.
        """
        super().__init__()
        self.kernel_size = kernel_size
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.codomain_elements = codomain_elements
        self.depthwise = convolution(
            kernel_size=kernel_size,
            in_channels=in_channels,
            groups=in_channels,
            out_channels=in_channels,
            bias=False,
        )  # type: ignore
        self.pointwise = convolution(
            kernel_size=1,
            in_channels=in_channels * self.depthwise.channel_multiplexing_factor,
            out_channels=out_channels,
            groups=1,
            bias=False,
        )
        depthwise_shape = (in_channels, kernel_size)

        def depthwise_input() -> Tensor:
            return create_codomain_for_depthwise_1d_conv(
                depthwise_shape, self.codomain_elements
            )

        depthwise_precomputable_tag = precomputable(
            input_shape=depthwise_shape, input_generator=depthwise_input
        )
        self.depthwise = depthwise_precomputable_tag(self.depthwise)

        pointwise_shape = (
            1,
            in_channels * self.depthwise.channel_multiplexing_factor,
        )

        def pointwise_input():
            return create_codomain_for_1d_conv(
                pointwise_shape, self.depthwise.codomain_elements
            )

        pointwise_precomputable_tag = precomputable(
            input_shape=pointwise_shape, input_generator=pointwise_input
        )
        self.pointwise = pointwise_precomputable_tag(self.pointwise)

    def forward(self, x):
        return self.pointwise(self.depthwise(x))


def define_split_convolution(
    conv_cls, _channel_multiplexing_factor, input_domain_elements
):
    def wrapper(cls):
        class Wrapped(SplitConvolutionBase):
            channel_multiplexing_factor = _channel_multiplexing_factor

            def __init__(self, kernel_size, in_channels, out_channels):
                super().__init__(
                    convolution=conv_cls,
                    kernel_size=kernel_size,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    codomain_elements=input_domain_elements,
                )

        return Wrapped

    return wrapper


@define_split_convolution(
    BinaryActivatedConv1d, _channel_multiplexing_factor=1, input_domain_elements=[-1, 1]
)
class BinarySplitConv:
    pass


@define_split_convolution(
    TernaryConvolution1d,
    _channel_multiplexing_factor=1,
    input_domain_elements=[-1, 0, 1],
)
class TernarySplitConv:
    pass


@define_split_convolution(
    MultilevelResidualBinarizationConv1d,
    _channel_multiplexing_factor=2,
    input_domain_elements=[-1, 1],
)
class TwoBitSplitConv:
    pass



import unittest
from itertools import filterfalse

from elasticai.creator import resource_utils
from elasticai.creator.vhdl.language import Code


class VHDLFileTestCase(unittest.TestCase):
    def compareToFile(self, vhdl_file: str, generated_code: Code):
        vhdl_code = resource_utils.read_text(
            "elasticai.creator.integrationTests.vhdl", vhdl_file
        )

        def line_is_empty(line):
            return len(line) == 0

        vhdl_code = filterfalse(line_is_empty, map(str.strip, vhdl_code.splitlines()))

        def starts_with_comment(line):
            return line.startswith("--")

        vhdl_code = filterfalse(starts_with_comment, vhdl_code)
        vhdl_code = list(vhdl_code)
        generated_code = list(filterfalse(starts_with_comment, generated_code))
        self.assertEqual(vhdl_code, generated_code)



from itertools import repeat
from string import Template
from typing import Iterator, Sequence


def expand_multiline_template(template: str, **kwargs: Sequence) -> Iterator[str]:
    """Expand a template field to multiple lines, while keeping indentation.
    Example:
        >>> template = "\\t$my_key"
        >>> values = ["hello,", "world", "!"]
        >>> "\\n".join(expand_multiline_template(template, my_key=values))
        '\\thello,\\n\\tworld\\n\\t!'
    """
    lines = template.splitlines()
    for line in lines:
        contains_no_key = True
        for key in kwargs:
            if f"${key}" in line:
                contains_no_key = False
                for placeholder_line, value in zip(repeat(line), kwargs[key]):
                    t = Template(placeholder_line)
                    yield t.safe_substitute({key: value})
                break
        if contains_no_key:
            yield line



import torch
from torch.nn import Conv1d
from torch.nn.utils.parametrize import register_parametrization

from elasticai.creator.qat.layers import Binarize

layer = Conv1d(in_channels=2, out_channels=3, kernel_size=(1,), bias=False)

register_parametrization(layer, "weight", Binarize())

print(layer)
x = torch.tensor([[[1], [1]]], dtype=torch.float32)
y = layer(x)
print(y)



# -*- coding: utf-8 -*-
"""Basic Qtorch application"""
from functools import partial

import torch
import torchvision
from torch import nn, optim
from torch.nn import functional as F
from torch.utils.data import DataLoader, TensorDataset

from elasticai.creator.qat.constraints import WeightClipper
from elasticai.creator.qat.layers import Binarize, QConv2d, QLinear

xy_train = torchvision.datasets.FashionMNIST(
    root="data/",
    download=True,
)

(x_train, y_train) = xy_train.data, xy_train.targets
x_train = x_train.float() / 255.0
mean = x_train.mean()
std = x_train.std()
x_train = (x_train - mean) / std
xy_train = TensorDataset(x_train, y_train)

xy_valid = torchvision.datasets.FashionMNIST(
    train=False,
    root="data/",
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(mean=(mean), std=(std)),
        ]
    ),
    download=True,
)

print("mean: ", mean)


class Lambda(nn.Module):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def forward(self, x):
        return self.func(x)


dev = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


convolutions = []


def Conv2d(**kwargs):
    layer = QConv2d(**kwargs)
    convolutions.append(layer)
    return layer


def Linear(**kwargs):
    layer = QLinear(**kwargs)
    convolutions.append(layer)
    return layer


def get_model():
    _Conv2d = partial(Conv2d, quantizer=Binarize(), bias=False)
    # _Conv2d = nn.Conv2d
    num_bits = 1
    activation = Binarize
    fc = partial(
        Linear, quantizer=Binarize(), bias=False, constraints=[WeightClipper()]
    )
    # fc = nn.Linear
    model = nn.Sequential(
        Lambda(lambda x: x.view(-1, 1, 28, 28)),
        _Conv2d(
            in_channels=1,
            out_channels=32,
            kernel_size=3,
            stride=2,
            padding=1,
            constraints=[WeightClipper()],
        ),
        nn.BatchNorm2d(32),
        activation(),
        _Conv2d(
            in_channels=num_bits * 32,
            out_channels=32,
            kernel_size=3,
            stride=2,
            padding=1,
            constraints=[WeightClipper()],
        ),
        nn.BatchNorm2d(32),
        activation(),
        _Conv2d(
            in_channels=num_bits * 32,
            out_channels=16,
            kernel_size=3,
            stride=2,
            padding=1,
            constraints=[WeightClipper()],
        ),
        nn.BatchNorm2d(16),
        activation(),
        Lambda(lambda x: x.view(x.size(0), -1)),
        fc(in_features=256 * num_bits, out_features=10),
    )
    return model, optim.Adamax(model.parameters(), lr=0.001)


model, optimizer = get_model()


def accuracy(out, yb):
    preds = torch.argmax(out, dim=1)
    return (preds == yb).float().mean()


def apply_weights_constraint():
    with torch.no_grad():
        for module in model.modules():
            if hasattr(module, "constraints"):
                module.apply_constraint()


def fit(model, optimizer, loss_func, batch_size, epochs, train_dataset, valid_dataset):
    data_loader = DataLoader(train_dataset, batch_size=batch_size)
    valid_loader = DataLoader(valid_dataset, batch_size=batch_size * 2)
    for epoch in range(epochs):
        model.train()
        for xb, yb in data_loader:
            pred = model(xb)
            loss = loss_func(pred, yb)
            loss.backward()
            optimizer.step()
            apply_weights_constraint()
            optimizer.zero_grad()
        model.eval()
        with torch.no_grad():
            valid_loss = sum(loss_func(model(xb), yb) for xb, yb in valid_loader)
            valid_accuracy = sum(accuracy(model(xb), yb) for xb, yb in valid_loader)
            train_accuracy = sum(accuracy(model(xb), yb) for xb, yb in data_loader)

        print(
            epoch,
            " :",
            valid_loss / len(valid_loader),
            ", ",
            valid_accuracy / len(valid_loader),
            " train acc: ",
            train_accuracy / len(data_loader),
        )


fit(
    model=model,
    optimizer=optimizer,
    loss_func=F.cross_entropy,
    batch_size=256,
    epochs=30,
    train_dataset=xy_train,
    valid_dataset=xy_valid,
)



from typing import Iterator

import numpy as np
from numpy.typing import NDArray


class IOTable:
    def __init__(self, inputs: np.ndarray, outputs: np.ndarray):
        self.inputs = inputs
        self.outputs = outputs

    def __getitem__(self, item):
        return self.inputs[item], self.outputs[item]

    def __len__(self):
        return len(self.inputs)

    def __iter__(self) -> Iterator[tuple]:
        for i in range(len(self)):
            yield self[i]

    def get_table_as_dict(self) -> dict:
        """given a list or single input,output table pair will return a list of dictionaries for each io pair.
            Said tables will be flatenned.
        Args:
            tables: io tables in the format
        Returns:
          dict: A list of dictionaries for the io_tables both ends will be flattened and transformed into tuples to
                facilitate iterating.
        """

        def to_tuple(x: NDArray[np.float_ | np.int_]) -> tuple[float | int, ...]:
            # noinspection PyTypeChecker
            native_python: list[float | int] = x.flatten().tolist()
            return tuple(native_python)

        return {to_tuple(x): to_tuple(y) for x, y in self}

    def grouped(self, groups: int) -> Iterator["IOTable"]:
        """given an input, output pair return a list of arrays describing the io Tables of each group.
        Args:
            table: IOTable the target IOTable
            groups: int the number of groups that will separate the said tables, those groups will break the tables along the first non-batch dimension.
        Returns:
          list[IOTable]: The grouped tables as a list of input output objects
        """
        inputs = self.inputs
        outputs = self.outputs
        assert (
            inputs.shape[1] % groups == 0 & outputs.shape[1] % groups == 0
        ), "the first dimension of the arrays should be divisible by the number of groups"
        inputs, outputs = inputs, outputs
        input_io_dim = list(inputs.shape)
        output_io_dim = list(outputs.shape)
        input_io_dim[1] = input_io_dim[1] // groups
        output_io_dim[1] = output_io_dim[1] // groups
        io_tables = [
            (np.zeros(input_io_dim), np.zeros(output_io_dim)) for _ in range(groups)
        ]
        for num, (input, output) in enumerate(zip(inputs, outputs)):
            for group in range(groups):
                input_length = input.shape[0] // groups
                output_length = output.shape[0] // groups
                io_tables[group][0][num] = input[
                    group * input_length : group * input_length + input_length
                ]
                io_tables[group][1][num] = output[
                    group * output_length : group * output_length + output_length
                ]
        for table in io_tables:
            yield IOTable(table[0], table[1])



"""
We perform a simple part of speech tagging using a quantized lstm network.
We proceed analogue to https://pytorch.org/tutorials/beginner/nlp/sequence_models_tutorial.html#example-an-lstm-for-part-of-speech-tagging
"""
import torch
from torch import nn, optim
from torch.nn import functional as F

from elasticai.creator.qat.layers import QLSTM, Binarize


def prepare_sequence(seq, to_ix):
    idxs = [to_ix[w] for w in seq]
    return torch.tensor(idxs, dtype=torch.long)


training_data = [
    # Tags are: DET - determiner; NN - noun; V - verb
    # For example, the word "The" is a determiner
    ("The dog ate the apple".split(), ["DET", "NN", "V", "DET", "NN"]),
    ("Everybody read that book".split(), ["NN", "V", "DET", "NN"]),
]
word_to_ix = {}
# For each words-list (sentence) and tags-list in each tuple of training_data
for sent, tags in training_data:
    for word in sent:
        if word not in word_to_ix:  # word has not been assigned an index yet
            word_to_ix[word] = len(word_to_ix)  # Assign each word with a unique index
tag_to_ix = {"DET": 0, "NN": 1, "V": 2}  # Assign each tag with a unique index

# These will usually be more like 32 or 64 dimensional.
# We will keep them small, so we can see how the weights change as we train.
EMBEDDING_DIM = 6
HIDDEN_DIM = 6


class LSTMTagger(nn.Module):
    def __init__(self, embedding_dim, hidden_dim, vocab_size, tagset_size):
        super(LSTMTagger, self).__init__()
        self.hidden_dim = hidden_dim

        self.word_embeddings = nn.Embedding(vocab_size, embedding_dim)

        # The LSTM takes word embeddings as inputs, and outputs hidden states
        # with dimensionality hidden_dim.
        self.lstm = QLSTM(
            embedding_dim,
            hidden_dim,
            state_quantizer=Binarize(),
            weight_quantizer=Binarize(),
        )

        self.binarize = Binarize()

        # The linear layer that maps from hidden state space to tag space
        self.hidden2tag = nn.Linear(hidden_dim, tagset_size)

    def forward(self, sentence):
        embeds = self.word_embeddings(sentence)
        embeds = self.binarize(embeds)
        lstm_out, _ = self.lstm(embeds.view(len(sentence), 1, -1))
        lstm_out = self.binarize(lstm_out)
        tag_space = self.hidden2tag(lstm_out.view(len(sentence), -1))
        tag_scores = F.log_softmax(tag_space, dim=1)
        return tag_scores


model = LSTMTagger(EMBEDDING_DIM, HIDDEN_DIM, len(word_to_ix), len(tag_to_ix))
loss_function = nn.NLLLoss()
optimizer = optim.SGD(model.parameters(), lr=0.1)

# See what the scores are before training
# Note that element i,j of the output is the score for tag j for word i.
# Here we don't need to train, so the code is wrapped in torch.no_grad()
with torch.no_grad():
    inputs = prepare_sequence(training_data[0][0], word_to_ix)
    tag_scores = model(inputs)

for epoch in range(300):  # again, normally you would NOT do 300 epochs, it is toy data
    for sentence, tags in training_data:
        # Step 1. Remember that Pytorch accumulates gradients.
        # We need to clear them out before each instance
        model.zero_grad()

        # Step 2. Get our inputs ready for the network, that is, turn them into
        # Tensors of word indices.
        sentence_in = prepare_sequence(sentence, word_to_ix)
        targets = prepare_sequence(tags, tag_to_ix)

        # Step 3. Run our forward pass.
        tag_scores = model(sentence_in)

        # Step 4. Compute the loss, gradients, and update the parameters by
        #  calling optimizer.step()
        loss = loss_function(tag_scores, targets)
        loss.backward()
        optimizer.step()

# See what the scores are after training
with torch.no_grad():
    # The sentence is "the dog ate the apple".  i,j corresponds to score for tag j
    # for word i. The predicted tag is the maximum scoring tag.
    # Here, we can see the predicted sequence below is 0 1 2 0 1
    # since 0 is index of the maximum value of row 1,
    # 1 is the index of maximum value of row 2, etc.
    # Which is DET NOUN VERB DET NOUN, the correct sequence!
    expected_tags = {"The": "DET", "dog": "NN", "ate": "V", "the": "DET", "apple": "NN"}
    actual_tags = {}
    inputs = prepare_sequence(expected_tags.keys(), word_to_ix)
    tag_scores = model(inputs)

    for index, word in enumerate(training_data[0][0]):
        tag = torch.argmax(tag_scores[index])
        tag = list(tag_to_ix.keys())[tag]
        actual_tags[word] = tag

    print("binarized embeddings: ", model.binarize(model.word_embeddings(inputs)))
    prediction_is_correct = expected_tags == actual_tags
    if not prediction_is_correct:
        print("predicted tags: ", actual_tags)
        print("but expected was: ", expected_tags)
    else:
        print("Test Successful!")



import math
from itertools import chain
from typing import Iterable, Iterator

import torch.nn

from elasticai.creator.vhdl.language import (
    Architecture,
    Code,
    CodeGenerator,
    ComponentDeclaration,
    ContextClause,
    DataType,
    Entity,
    InterfaceVariable,
    LibraryClause,
    PortMap,
    Process,
    UseClause,
    bin_representation,
)
from elasticai.creator.vhdl.language_testbench import TestBenchBase
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
    infer_total_and_frac_bits,
)


def _vhdl_add_assignment(
    code: list, line_id: str, value: str, comment: str = None
) -> None:
    new_code_fragment = f"{line_id} <= {bin_representation(value)};"
    if comment is not None:
        new_code_fragment += f" -- {comment}"
    code.append(new_code_fragment)


def precomputed_scalar_function_process(
    x: list[FixedPoint], y: list[FixedPoint]
) -> CodeGenerator:
    """
        returns the string of a lookup table
    Args:
        x (FixedPoint): input list
        y (FixedPoint) : output list
    Returns:
        String of lookup table (if/elsif statements for vhdl file)
    """
    x.sort()
    lines = []
    if len(x) == 0 and len(y) == 1:
        _vhdl_add_assignment(
            code=lines,
            line_id="y",
            value=y[0].to_bin(),
        )
    elif len(x) != len(y) - 1:
        raise ValueError(
            f"x has to be one element shorter than y, but x has {len(x)} elements and y {len(y)} elements"
        )
    else:
        smallest_possible_output = y[0]
        biggest_possible_output = y[-1]

        # first element
        for x_value in x[:1]:
            lines.append(f"if int_x<{x_value.to_signed_int()} then")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=smallest_possible_output.to_bin(),
                comment=str(smallest_possible_output.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        for current_x, current_y in zip(x[1:], y[1:-1]):
            lines.append(f"elsif int_x<{current_x.to_signed_int()} then")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=current_y.to_bin(),
                comment=str(current_y.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        # last element only in y
        for _ in y[-1:]:
            lines.append("else")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=biggest_possible_output.to_bin(),
                comment=str(biggest_possible_output.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        if len(lines) != 0:
            lines.append("end if;")

    # build the string block

    def generator() -> Code:
        for line in lines:
            yield line

    return generator


class DataWidthVariable(InterfaceVariable):
    def __init__(self, value: int):
        super().__init__(
            identifier="DATA_WIDTH", identifier_type=DataType.INTEGER, value=f"{value}"
        )


class FracWidthVariable(InterfaceVariable):
    def __init__(self, value: int):
        super().__init__(
            identifier="FRAC_WIDTH", identifier_type=DataType.INTEGER, value=f"{value}"
        )


class PrecomputedScalarFunction:
    def __init__(
        self,
        x: list[FixedPoint],
        y: list[FixedPoint],
        component_name: str = None,
        process_instance: Process = None,
    ):
        """
        We calculate the function with an algorithm equivalent to:
        ```
        def function(x: int, inputs: list[int], outputs: list[int]) -> int:
          for input, output in zip(inputs, outputs[:-1]):
            if x < input:
              return output
          return outputs[-1]
        ```
        """
        self.component_name = self._get_lower_case_class_name_or_component_name(
            component_name=component_name
        )

        self.data_width, self.frac_width = infer_total_and_frac_bits(x, y)
        self.x = x
        self.y = y
        self.process_instance = process_instance

    @classmethod
    def _get_lower_case_class_name_or_component_name(cls, component_name):
        if component_name is None:
            return cls.__name__.lower()
        return component_name

    @property
    def file_name(self) -> str:
        return f"{self.component_name}.vhd"

    def __call__(self) -> Iterable[str]:
        library = ContextClause(
            library_clause=LibraryClause(logical_name_list=["ieee"]),
            use_clause=UseClause(
                selected_names=[
                    "ieee.std_logic_1164.all",
                    "ieee.numeric_std.all",
                ]
            ),
        )
        entity = Entity(self.component_name)
        entity.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
        ]
        entity.port_list = [
            "x : in signed(DATA_WIDTH-1 downto 0)",
            "y : out signed(DATA_WIDTH-1 downto 0)",
        ]
        process = Process(
            identifier=self.component_name,
            lookup_table_generator_function=precomputed_scalar_function_process(
                x=self.x, y=self.y
            ),
            input_name="x",
        )
        process.process_declaration_list = ["variable int_x: integer := 0"]
        process.process_statements_list = ["int_x := to_integer(x)"]
        architecture = Architecture(
            design_unit=self.component_name,
        )
        architecture.architecture_statement_part = process
        code = chain(chain(library(), entity()), architecture())
        return code


class Sigmoid(PrecomputedScalarFunction):
    def __init__(self, x: list[FixedPoint], component_name: str = None):
        x_tensor = torch.as_tensor(list(map(float, x)))
        # calculate y always for the previous element, therefore the last input is not needed here
        y = torch.nn.Sigmoid()(x_tensor[:-1]).tolist()
        y.insert(0, 0)
        # add last y value, therefore, x_tensor is one element shorter than y_tensor
        y.append(1)
        y = float_values_to_fixed_point(y, *infer_total_and_frac_bits(x))

        super(Sigmoid, self).__init__(x=x, y=y, component_name=component_name)


class Tanh(PrecomputedScalarFunction):
    def __init__(self, x: list[FixedPoint], component_name: str = None):
        y_list = [-1.0]
        # calculate y always for the previous element, therefore the last input is not needed here
        for x_element in x[:-1]:
            y_list.append(math.tanh(float(x_element)))
        # add last y value, therefore, x_list is one element shorter than y_list
        y_list.append(1)
        y = float_values_to_fixed_point(y_list, *infer_total_and_frac_bits(x))

        super(Tanh, self).__init__(x=x, y=y, component_name=component_name)


class PrecomputedScalarTestBench:
    def __init__(
        self,
        x_list_for_testing: list[FixedPoint],
        y_list_for_testing: list[FixedPoint],
        component_name: str = None,
    ):
        self.component_name = self._get_lower_case_class_name_or_component_name(
            component_name=component_name
        )
        self.data_width, self.frac_width = infer_total_and_frac_bits(
            x_list_for_testing, y_list_for_testing
        )
        self.x_list_for_testing = x_list_for_testing
        self.y_list_for_testing = y_list_for_testing

    @classmethod
    def _get_lower_case_class_name_or_component_name(cls, component_name):
        if component_name is None:
            return cls.__name__.lower()
        return component_name

    @property
    def file_name(self) -> str:
        return f"{self.component_name}_tb.vhd"

    def __call__(self) -> Iterable[str]:
        library = ContextClause(
            library_clause=LibraryClause(logical_name_list=["ieee"]),
            use_clause=UseClause(
                selected_names=[
                    "ieee.std_logic_1164.all",
                    "ieee.numeric_std.all",
                    "ieee.math_real.all",
                ]
            ),
        )

        entity = Entity(self.component_name + "_tb")
        entity.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
        ]
        entity.port_list = [
            "clk : out std_logic",
        ]

        component = ComponentDeclaration(identifier=self.component_name)
        component.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
        ]
        component.port_list = [
            "x : in signed(DATA_WIDTH-1 downto 0)",
            "y : out signed(DATA_WIDTH-1 downto 0)",
        ]

        process = Process(
            identifier="clock",
        )
        process.process_statements_list = [
            "clk <= '0'",
            "wait for clk_period/2",
            "clk <= '1'",
            "wait for clk_period/2",
        ]

        uut_port_map = PortMap(map_name="uut", component_name=self.component_name)
        uut_port_map.signal_list.append("x => test_input")
        uut_port_map.signal_list.append("y => test_output")

        test_cases = TestCasesPrecomputedScalarFunction(
            x_list_for_testing=self.x_list_for_testing,
            y_list_for_testing=self.y_list_for_testing,
        )
        test_process = Process(identifier="test")
        test_process.process_statements_list = [t for t in test_cases()]

        architecture = Architecture(
            design_unit=self.component_name + "_tb",
        )
        architecture.architecture_declaration_list = [
            "signal clk_period : time := 1 ns",
            "signal test_input : signed(16-1 downto 0):=(others=>'0')",
            "signal test_output : signed(16-1 downto 0)",
        ]

        architecture.architecture_component_list.append(component)
        architecture.architecture_process_list.append(process)
        architecture.architecture_port_map_list.append(uut_port_map)
        architecture.architecture_statement_part = test_process

        code = chain(library(), entity(), architecture())
        return code


class TestCasesPrecomputedScalarFunction(TestBenchBase):
    def __init__(
        self,
        x_list_for_testing: list[FixedPoint],
        y_list_for_testing: list[FixedPoint],
        x_variable_name: str = "test_input",
        y_variable_name: str = "test_output",
    ):
        assert len(x_list_for_testing) == len(y_list_for_testing)
        self.data_width, _ = infer_total_and_frac_bits(
            x_list_for_testing, y_list_for_testing
        )
        self.x_list_for_testing = [
            value.to_signed_int() for value in x_list_for_testing
        ]
        self.y_list_for_testing = [
            value.to_signed_int() for value in y_list_for_testing
        ]
        self.x_variable_name = x_variable_name
        self.y_variable_name = y_variable_name

    def __len__(self):
        return len(self.y_list_for_testing)

    def _body(self) -> Iterator[str]:
        for x_value, y_value in zip(self.x_list_for_testing, self.y_list_for_testing):
            yield f"{self.x_variable_name} <= to_signed({x_value},{self.data_width})"
            yield f"wait for 1*clk_period"
            yield f"report \"The value of '{self.y_variable_name}' is \" & integer'image(to_integer(unsigned({self.y_variable_name})))"
            if isinstance(y_value, str):
                yield f'assert {self.y_variable_name}="{y_value}" report "The test case {x_value} fail" severity failure'
            else:
                yield f'assert {self.y_variable_name}={y_value} report "The test case {x_value} fail" severity failure'

    def __call__(self) -> Iterable[Code]:
        yield from iter(self)



import unittest

import torch
from torch.nn import Conv2d

from elasticai.creator.qat.QRigL import QRigLScheduler


class test_QRigl(unittest.TestCase):
    def test_QrigL_initialization_check_mask(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=3)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=2,
            T_end=0,
        )
        self.assertSequenceEqual(layer.weight.shape, scheduler.backward_masks[0].shape)
        self.assertEqual(torch.count_nonzero(scheduler.backward_masks[0][0]), 2)

    def test_QrigL_weight_is_initially_multiplied_by_mask(self):
        layer = Conv2d(in_channels=4, out_channels=2, kernel_size=3)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=2,
            T_end=0,
        )
        self.assertTrue(
            torch.all(layer.weight == layer.weight * scheduler.backward_masks[0])
        )

    def test_QrigL_weight_is_after_step_multiplied_by_mask(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=3)
        optimizer = torch.optim.Adam([layer.weight])
        scheduler = QRigLScheduler(
            weighted_layers=(layer,), optimizer=optimizer, num_connections=2, T_end=0
        )
        new_mask = torch.zeros_like(layer.weight)
        scheduler.backward_masks[0] = new_mask
        optimizer.step()
        self.assertTrue(torch.all(layer.weight == new_mask))

    def test_QrigL_check_grad_accumulation(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=1)
        optimizer = torch.optim.Adam([layer.weight])
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=optimizer,
            num_connections=4,
            delta=4,
            grad_accumulation_n=2,
            T_end=100,
        )
        for i in range(3):
            optimizer.zero_grad()
            test_input = torch.ones((1, 2, 1, 1))
            out = layer(test_input)
            out.mean().backward()
            scheduler()
        n = scheduler.backward_hook_objects[0].dense_grad
        m = layer.weight.grad
        self.assertTrue(
            torch.all(
                layer.weight.grad / 2 == scheduler.backward_hook_objects[0].dense_grad
            )
        )

    def Qrigl_step_helper(self):
        layer = Conv2d(in_channels=2, out_channels=1, kernel_size=1)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=1,
            T_end=100,
            alpha=1,
        )
        layer.weight.grad = torch.ones_like(layer.weight)
        scheduler.backward_hook_objects[0].dense_grad = torch.zeros_like(layer.weight)
        scheduler.backward_hook_objects[0].dense_grad[0][0] = 1
        scheduler.backward_hook_objects[0].dense_grad[0][1] = 0
        scheduler.backward_masks[0][0][0] = 0
        scheduler.backward_masks[0][0][1] = 1
        return layer, scheduler

    def test_QrigL_rigl_after_step_mask(self):
        layer, scheduler = self.Qrigl_step_helper()
        scheduler._rigl_step()
        self.assertEqual(scheduler.backward_masks[0][0][0][0][0].item(), 1.0)
        self.assertEqual(scheduler.backward_masks[0][0][1][0][0].item(), 0.0)

    def test_QrigL_grad_is_after_step_multiplied_by_mask(self):
        layer, scheduler = self.Qrigl_step_helper()
        scheduler._rigl_step()
        expected_grad = torch.ones_like(layer.weight)
        expected_grad[0][1] = 0
        self.assertTrue(torch.all(layer.weight.grad == expected_grad))

    def test_QrigL_call(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=3)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=2,
            T_end=101,
            delta=10,
        )
        for call in range(200):
            layer.weight.grad = torch.zeros_like(layer.weight)
            scheduler()
        self.assertEqual(scheduler.step, 200)
        self.assertEqual(scheduler.rigl_steps, 10)

    def test_QrigL_momentum(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=1)
        optimizer = torch.optim.SGD([layer.weight], lr=0.01, momentum=0.1)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=optimizer,
            num_connections=2,
            T_end=101,
            delta=10,
        )
        scheduler.backward_masks[0] = torch.zeros_like(layer.weight)
        scheduler.backward_masks[0][0][1] = 1
        for i in range(3):
            optimizer.zero_grad()
            test_input = torch.ones((1, 2, 1, 1))
            out = layer(test_input)
            out.mean().backward()
            scheduler()
            optimizer.step()
        buffer = optimizer.state[layer.weight]["momentum_buffer"]
        self.assertTrue(torch.all(buffer == buffer * scheduler.backward_masks[0]))

    def test_QrigL_state_dict(self):
        layer = Conv2d(in_channels=2, out_channels=2, kernel_size=3)
        scheduler = QRigLScheduler(
            weighted_layers=(layer,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=2,
            T_end=101,
            delta=10,
        )
        layer2 = Conv2d(in_channels=2, out_channels=2, kernel_size=3)
        scheduler2 = QRigLScheduler(
            weighted_layers=(layer2,),
            optimizer=torch.optim.Adam([layer.weight]),
            num_connections=0,
            T_end=0,
            delta=0,
            state_dict=scheduler.state_dict(),
        )
        a = scheduler == scheduler2
        self.assertTrue(scheduler == scheduler2)


if __name__ == "__main__":
    unittest.main()



from types import SimpleNamespace

from elasticai.creator.tags_utils import get_tags, tag
from elasticai.creator.tests.tensor_test_case import TensorTestCase


class TagTest(TensorTestCase):
    def test_precompute_block(self):
        module = SimpleNamespace()
        module = tag(module, precomputed=None)
        self.assertIn("precomputed", module.elasticai_tags.keys())

    def test_precomputed_blocks_children_are_accessible(self):
        def children_generator():
            yield from [1, 2]

        module = SimpleNamespace(children=children_generator)
        module = tag(module)
        self.assertSequenceEqual([1, 2], list(module.children()))

    def test_precomputed_blocks_other_attributes_are_accessible(self):
        module = SimpleNamespace(other_attribute="some other attribute")
        module = tag(module)
        self.assertEqual("some other attribute", module.other_attribute)

    def test_update_tags(self):
        module = SimpleNamespace()
        module = tag(module, precomputed=None)
        module = tag(module, input_tensor_shape=(1, 2, 3))
        self.assertEqual(
            {
                "precomputed": None,
                "input_tensor_shape": (1, 2, 3),
            },
            get_tags(module),
        )



from functools import partial

from elasticai.creator.integrationTests.vhdl.vhdl_file_test_case import VHDLFileTestCase
from elasticai.creator.vhdl.lstm_testbench_generator import LSTMCellTestBench
from elasticai.creator.vhdl.number_representations import (
    float_values_to_fixed_point,
    int_values_to_fixed_point,
)


class LSTMCellTestBenchTest(VHDLFileTestCase):
    maxDiff = None

    def test_compare_files(self) -> None:
        fp_args = dict(total_bits=16, frac_bits=8)
        ints_to_fp = partial(int_values_to_fixed_point, **fp_args)
        floats_to_fp = partial(float_values_to_fixed_point, **fp_args)

        lstm_cell = LSTMCellTestBench(
            input_size=5,
            hidden_size=20,
            component_name="lstm_cell",
            test_x_h_data=ints_to_fp(
                [
                    0x018A,
                    0xFFB5,
                    0xFDD3,
                    0x0091,
                    0xFEEB,
                    0x0099,
                    0xFE72,
                    0xFFA9,
                    0x01DA,
                    0xFFC9,
                    0xFF42,
                    0x0090,
                    0x0042,
                    0xFFD4,
                    0xFF53,
                    0x00F0,
                    0x007D,
                    0x0134,
                    0x0015,
                    0xFECD,
                    0xFFFF,
                    0xFF7C,
                    0xFFB2,
                    0xFE6C,
                    0x01B4,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                ]
            ),
            test_c_data=ints_to_fp(
                [
                    0x0034,
                    0xFF8D,
                    0xFF6E,
                    0xFF72,
                    0xFEE0,
                    0xFFAF,
                    0xFEE9,
                    0xFFEB,
                    0xFFE9,
                    0x00AF,
                    0xFF2A,
                    0x0000,
                    0xFF40,
                    0x002F,
                    0x009F,
                    0x00A3,
                    0xFFC2,
                    0x024D,
                    0xFE1F,
                    0xFFF4,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                    0x0000,
                ]
            ),
            h_out=floats_to_fp(
                [
                    34,
                    -80,
                    -32,
                    -28,
                    -88,
                    11,
                    -60,
                    6,
                    -16,
                    18,
                    -32,
                    46,
                    -77,
                    15,
                    70,
                    27,
                    13,
                    112,
                    -126,
                    3,
                ]
            ),
        )
        lstm_cell_code = lstm_cell()
        self.compareToFile("expected_lstm_cell_testbench.vhd", lstm_cell_code)



import matplotlib.pyplot as plt
import torch
from torch.utils.data import DataLoader, Dataset

from elasticai.creator.qat.layers import QLSTM, QLinear


class SinusDataset(Dataset):
    def __init__(self, sinus_range=(1, 500), num_points=5000, window_size=100):
        self.sinus_range = sinus_range
        self.num_points = num_points
        self.window_size = window_size

        x = torch.linspace(*self.sinus_range, self.num_points)
        self._full_series = (
            torch.sin(x) * (self.sinus_range[1] - x) / self.sinus_range[1]
        )
        self._data = self._rolling_samples_with_labels(
            self._full_series, self.window_size
        )
        self._data = self._reshape_samples(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, idx):
        return self._data[idx]

    @staticmethod
    def _rolling_samples_with_labels(data, window_size):
        result = []
        for i in range(window_size, len(data)):
            result.append((data[i - window_size : i], data[i]))
        return result

    @staticmethod
    def _reshape_samples(data):
        return [(sample.view(-1, 1), label) for sample, label in data]

    @property
    def full_series(self):
        return self._full_series


class QLSTMModel(torch.nn.Module):
    def __init__(self, window_size):
        super().__init__()
        # NOTE: With quantization I get really high loss values. But I don't spend lot of time in quantizing this model
        self.lstm1 = QLSTM(
            input_size=1,
            hidden_size=64,
            bias=True,  # bias=False
            state_quantizer=None,  # state_quantizer=Binarize()
            weight_quantizer=None,  # weight_quantizer=Binarize()
            input_gate_activation=None,  # input_gate_activation=Binarize()
            forget_gate_activation=None,  # forget_gate_activation=Binarize()
            cell_gate_activation=None,  # cell_gate_activation=Ternarize()
            output_gate_activation=None,  # output_gate_activation=Binarize()
            new_cell_state_activation=None,  # new_cell_state_activation=Ternarize()
            batch_first=True,
        )
        #        self.bin = Binarize()
        self.linear = QLinear(
            in_features=window_size * 64,
            out_features=1,
            bias=True,  # bias=False
            quantizer=None,  # quantizer=Binarize
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        x, _ = self.lstm1(inputs)
        #        x = self.bin(x)
        x = self.linear(x.view(len(x), -1))
        return x


if __name__ == "__main__":
    ds = SinusDataset(sinus_range=(1, 500), num_points=5000, window_size=100)

    # plt.plot(ds.full_series)
    # plt.show()

    num_train_samples = int(len(ds) * 0.8)
    ds_train = ds[:num_train_samples]
    ds_test = ds[num_train_samples:]

    batch_size = 32
    dl_train = DataLoader(ds_train, batch_size=batch_size, shuffle=True)
    dl_test = DataLoader(ds_test, batch_size=batch_size, shuffle=True)

    print("Number of train batches:", len(dl_train))
    print("Number of test batches:", len(dl_test))

    lstm_model = QLSTMModel(ds.window_size)

    n_epochs = 10
    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(lstm_model.parameters(), lr=1e-4)

    for epoch in range(n_epochs):
        losses = []

        for X, y in dl_train:
            y_pred = lstm_model(X)
            loss = loss_fn(
                y_pred.view(
                    -1,
                ),
                y,
            )
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            losses.append(loss.item())
        train_loss = sum(losses) / len(losses)

        with torch.no_grad():
            val_loss = sum(
                loss_fn(
                    lstm_model(X).view(
                        -1,
                    ),
                    y,
                )
                for X, y in dl_test
            ) / len(dl_test)

        print(
            "Epoch: {}/{}; Loss: {}; Val_Loss: {}".format(
                epoch + 1, n_epochs, train_loss, val_loss
            )
        )

    # Plot final result for one window
    idx = 1
    ws = ds.window_size
    sample = ds_test[idx][0].tolist()
    predicted = ds_test[idx][0].tolist()
    target = [target_value.item() for _, target_value in ds_test[idx : idx + ws]]

    for i in range(ws):
        y_pred = lstm_model(torch.tensor([predicted]))
        predicted = predicted[1:] + [
            y_pred.view(
                -1,
            ).tolist()
        ]
    x_axis = torch.arange(0, ws * 2)
    plt.plot(
        x_axis[:ws],
        torch.tensor(sample).view(
            -1,
        ),
        "-b",
        label="previous window",
    )
    plt.plot(
        x_axis[ws:],
        torch.tensor(target).view(
            -1,
        ),
        "-g",
        label="target prediction",
    )
    plt.plot(
        x_axis[ws:],
        torch.tensor(predicted).view(
            -1,
        ),
        "-r",
        label="actual prediction",
    )
    plt.legend()
    plt.show()



import itertools
import unittest

import numpy as np
import torch
from numpy.testing import assert_equal as assertNPArrayEqual

from elasticai.creator.precomputation.input_domains import (
    construct_codomain_from_elements,
    create_input_data,
    create_io_table,
    find_unique_elements,
    get_cartesian_product_from_items,
)


def PyTorchAdapter():
    return torch


class DeriveDataSetsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ml_lib = PyTorchAdapter()

    def test_create_input_data_1d(self) -> None:
        table = create_input_data(input_dim=[2], domain=[0, 1])
        self.assertTrue(table.tolist(), [[0, 0], [0, 1], [1, 0], [1, 1]])

    def test_create_input_data_none(self) -> None:
        table = create_input_data([2, None], [0, 1])
        self.assertTrue((table == [[0, 0], [0, 1], [1, 0], [1, 1]]).all())

    def test_create_input_data_2d(self) -> None:
        table = create_input_data([1, 2], [0, 1])
        self.assertTrue((table == [[[0, 0]], [[0, 1]], [[1, 0]], [[1, 1]]]).all())

    def test_create_input_data_ternary(self) -> None:
        table = create_input_data([1], [0, 1, -1])
        self.assertTrue((table == [[-1], [0], [1]]).all())

    def test_create_input_data_raiseSystemExit(self) -> None:
        self.assertRaises(MemoryError, create_input_data, [6, 6], [-1, 1])

    def test_create_io_table_basic(self) -> None:
        inputs = np.array([[1], [2], [3]])
        outputs = [
            self.ml_lib.tensor([2]),
            self.ml_lib.tensor([3]),
            self.ml_lib.tensor([4]),
        ]
        dict = create_io_table(inputs, outputs, channel_wise=False)
        self.assertTrue((dict == {(1,): (2,), (2,): (3,), (3,): (4,)}))

    def test_create_io_table_array(self) -> None:
        inputs = np.array([[1, 1], [2, 2], [3, 3]])
        outputs = [
            self.ml_lib.tensor([2, 2]),
            self.ml_lib.tensor([3, 3]),
            self.ml_lib.tensor([4, 4]),
        ]
        dict = create_io_table(inputs, outputs, channel_wise=False)
        self.assertTrue((dict == {(1, 1): (2, 2), (2, 2): (3, 3), (3, 3): (4, 4)}))

    def test_create_io_table_complex(self) -> None:
        inputs = np.array([[[1, 1], [1, 1]], [[2, 2], [2, 2]], [[3, 3], [3, 3]]])
        outputs = [
            self.ml_lib.tensor([2, 2]),
            self.ml_lib.tensor([3, 3]),
            self.ml_lib.tensor([4, 4]),
        ]
        dict = create_io_table(inputs, outputs, channel_wise=False)
        self.assertTrue(
            (dict == {(1, 1, 1, 1): (2, 2), (2, 2, 2, 2): (3, 3), (3, 3, 3, 3): (4, 4)})
        )

    def test_create_io_table_channel_wise_basic(self) -> None:
        inputs = np.transpose(np.array([[[1], [2]], [[3], [4]]]), (0, 2, 1))
        outputs = [self.ml_lib.tensor([[[5], [6]]]), self.ml_lib.tensor([[[7], [8]]])]
        io_table = create_io_table(inputs, outputs, channel_wise=True)
        self.assertTrue(
            (io_table == [{(1,): (5,), (3,): (7,)}, {(2,): (6,), (4,): (8,)}])
        )

    def test_create_io_table_channel_wise_array(self) -> None:
        inputs = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        outputs = [
            self.ml_lib.tensor([[2, 3], [4, 5]]),
            self.ml_lib.tensor([[6, 7], [8, 9]]),
        ]
        io_table = create_io_table(inputs, outputs, channel_wise=True)
        self.assertTrue(
            (
                io_table
                == [{(1, 3): (2, 4), (5, 7): (6, 8)}, {(2, 4): (3, 5), (6, 8): (7, 9)}]
            )
        )

    def test_find_unique_elements_basic(self) -> None:
        arr = find_unique_elements(np.array([0, 0, 1, 2]))
        self.assertTrue((arr == [0, 1, 2]).all())

    def test_find_unique_elements_complex(self) -> None:
        arr = find_unique_elements(
            np.array([[[1, 1], [2, 1]], [[2, 1], [1, 1]], [[1, 1], [2, 1]]])
        )
        self.assertTrue((list(arr.shape) == [2, 2, 2]))
        self.assertTrue(
            (
                np.ndarray.flatten(arr)
                == np.ndarray.flatten(np.array([[[1, 1], [2, 1]], [[2, 1], [1, 1]]]))
            ).all()
        )

    def test_cartesian_product_of_one_element_is_a_repetition(self) -> None:
        expected = np.array(((1, 1),), dtype="float16")
        actual = get_cartesian_product_from_items(length=2, items=[1])
        assertNPArrayEqual(actual, expected)

    def test_cartesian_product_of_one_element_is_repeated_three_times(self) -> None:
        expected = np.array(((1, 1, 1),), dtype="float16")
        actual = get_cartesian_product_from_items(length=3, items=[1])
        assertNPArrayEqual(actual, expected)

    def test_cartesian_product_for_two_elements(self):
        expected = np.array(((1, 1), (1, 2), (2, 1), (2, 2)), dtype="float16")
        actual = get_cartesian_product_from_items(length=2, items=(1, 2))
        assertNPArrayEqual(actual, expected)

    def test_cartesian_product_for_two_vector_elements(self):
        expected = (
            ((0, 0), (0, 0)),
            ((0, 0), (1, 1)),
            ((1, 1), (0, 0)),
            ((1, 1), (1, 1)),
        )
        expected = np.array(expected, dtype="float16")
        actual = get_cartesian_product_from_items(length=2, items=((0, 0), (1, 1)))
        assertNPArrayEqual(expected, actual)

    def test_cartesian_nested(self):
        expected = itertools.product((0, 1), repeat=4)
        expected = ([[a, b], [c, d]] for a, b, c, d in expected)
        expected = np.array(list(expected), dtype="float16")
        actual = get_cartesian_product_from_items(items=(0, 1), length=2)
        actual = get_cartesian_product_from_items(items=actual, length=2)
        assertNPArrayEqual(expected, actual)

    def test_construct_domain_yields_domain_with_2_2_shaped_elements(self):
        expected = itertools.product((0, 1), repeat=4)
        expected = ([[a, b], [c, d]] for a, b, c, d in expected)
        expected = np.array(list(expected), dtype="float16")
        actual = construct_codomain_from_elements(
            shape=(2, 2), codomain_elements=(0, 1)
        )
        assertNPArrayEqual(expected, actual)

    def test_construct_domain_from_subshape_vectors_length_2(self):
        expected = itertools.product((0, 1), repeat=2)
        table = ((0, 0), (1, 1))
        expected = ((table[i], table[j]) for i, j in expected)
        expected = np.array(list(expected), dtype="float16")
        actual = construct_codomain_from_elements(
            shape=(2, 2), codomain_elements=((0, 0), (1, 1))
        )
        assertNPArrayEqual(expected, actual)

    def test_construct_domain_from_subshape_vectors_length_3(self):
        expected = itertools.product((0, 1), repeat=3)
        table = ((0, 0), (1, 1))
        expected = ((table[i], table[j], table[k]) for i, j, k in expected)
        expected = np.array(list(expected), dtype="float16")
        actual = construct_codomain_from_elements(
            shape=(3, 2), codomain_elements=((0, 0), (1, 1))
        )
        assertNPArrayEqual(expected, actual)

    def test_construct_domain_from_incompatible_subshape_raises_error(self):
        self.assertRaises(
            ValueError,
            construct_codomain_from_elements,
            shape=(2, 3),
            codomain_elements=((0, 0), (1, 1)),
        )


if __name__ == "__main__":
    unittest.main()



""" Original: version:https://github.com/nollied/rigl-torch """
import numpy as np
import torch

from elasticai.creator.qat.masks import random_mask_4d


class IndexMaskHook:
    def __init__(self, layer, scheduler):
        self.layer = layer
        self.scheduler = scheduler
        self.dense_grad = None

    def __name__(self):
        return "IndexMaskHook"

    def __eq__(self, other):
        a = isinstance(other, IndexMaskHook) & (
            all(self.layer.shape == other.layer.shape)
        )
        return isinstance(other, IndexMaskHook) & (
            all(self.layer.shape == other.layer.shape)
        )

    @torch.no_grad()
    def __call__(self, grad):
        mask = self.scheduler.backward_masks[self.layer]

        # only calculate dense_grads when necessary
        if self.scheduler.check_if_backward_hook_should_accumulate_grad():
            if self.dense_grad is None:
                # initialize as all 0s so we can do a rolling average
                self.dense_grad = torch.zeros_like(grad)
            self.dense_grad += grad / self.scheduler.grad_accumulation_n
        else:
            self.dense_grad = None

        return grad * mask


def _create_step_wrapper(scheduler, optimizer):
    _unwrapped_step = optimizer.step

    def _wrapped_step():
        _unwrapped_step()
        scheduler.reset_momentum()
        scheduler.apply_mask_to_weights()

    optimizer.step = _wrapped_step


class QRigLScheduler:
    """
    Creates a Rigl scheduler for the selected Conv2d layers. This works similarly to the
    original Rigl [Evci et al 2021] https://arxiv.org/pdf/1911.11134.pdf , with the difference that it has a fixed
    number of connections per channel rather than a per layer fan in. This works by changing the connections,
    by editing a mask every few steps .This is useful in precalculated applications. It replaces the optimizer step
    every few steps. So it should be called like:
    if scheduler():
        optimizer.step()
    Args:
        layers. The Conv2d layers
        optimizer: a torch optimizer it will be wrapped by a function
        num connections: the fixed number of connections per channel
        delta: how often the update step should be done
        alpha: times cosine annealing determines which fraction of weights should get shifted.
        T_end: when the mask should be fixed, given by a fraction of total batches   default expected usage: 0.75 * epochs * len(trainLoader)
        grad_accumulation: added by the Rigl_torch author, takes the averaging score of n batches when determining the growth scores
    """

    def __init__(
        self,
        weighted_layers: list[torch.nn.Module],
        optimizer: torch.optim.Optimizer,
        num_connections: int,
        T_end,
        delta=100,
        alpha=0.3,
        grad_accumulation_n=1,
        state_dict=None,
    ):

        self.layers = weighted_layers
        self.optimizer = optimizer
        self.backward_masks = []
        _create_step_wrapper(self, optimizer)
        self.weights_in_layer = [layer.weight.numel() for layer in self.layers]
        self.num_connections = num_connections
        if state_dict is not None:
            self.load_state_dict(state_dict)
            self.apply_mask_to_weights()

        else:

            # randomly sparsify model according to S
            self.random_sparsify()

            assert 0 < grad_accumulation_n < delta
            # scheduler keeps a log of how many times it's called. this is how it does its scheduling
            self.step = 0
            self.rigl_steps = 0
            # define the actual schedule
            self.delta = delta
            self.alpha = alpha
            self.T_end = T_end
            self.grad_accumulation_n = grad_accumulation_n
        self.backward_hook_objects = []
        for i, layer in enumerate(self.layers):
            if getattr(layer.weight, "_has_rigl_backward_hook", False):
                raise Exception(
                    "This model already has been registered to a RigLScheduler."
                )

            self.backward_hook_objects.append(IndexMaskHook(i, self))
            layer.weight.register_hook(self.backward_hook_objects[-1])
            setattr(layer.weight, "_has_rigl_backward_hook", True)
        self.apply_mask_to_weights()

    def state_dict(self):
        obj = {
            "num_connections": self.num_connections,
            "weights_in_layer": self.weights_in_layer,
            "delta": self.delta,
            "alpha": self.alpha,
            "T_end": self.T_end,
            "grad_accumulation_n": self.grad_accumulation_n,
            "step": self.step,
            "rigl_steps": self.rigl_steps,
            "backward_masks": self.backward_masks,
        }

        return obj

    def __eq__(self, other):
        return (
            isinstance(other, QRigLScheduler)
            & (self.num_connections == other.num_connections)
            & (self.delta == other.delta)
            & (self.T_end == other.T_end)
            & (self.alpha == other.alpha)
            & (self.grad_accumulation_n == other.grad_accumulation_n)
            & (self.weights_in_layer == other.weights_in_layer)
            & (
                torch.all(
                    list(
                        a == b
                        for a, b in zip(self.backward_masks, other.backward_masks)
                    )[0]
                ).item()
            )
        )

    def load_state_dict(self, state_dict):
        for k, v in state_dict.items():
            if type(v) == dict:
                self.load_state_dict(v)
            setattr(self, k, v)

    @torch.no_grad()
    def random_sparsify(self):
        self.backward_masks = []
        for layer_index, layer in enumerate(self.layers):
            # if sparsity is 0%, skip

            mask = random_mask_4d(
                out_channels=layer.out_channels,
                kernel_size=layer.kernel_size,
                in_channels=layer.in_channels,
                groups=layer.groups,
                params_per_channel=self.num_connections,
            )
            layer.weight *= mask
            self.backward_masks.append(mask)

    def __str__(self):
        s = "RigLScheduler(\n"
        s += "layers=%i,\n" % len(self.weights_in_layer)

        # calculate the number of non-zero elements out of the total number of elements
        N_str = "["
        S_str = "["
        sparsity_percentages = []
        total_params = 0
        total_conv_params = 0
        total_nonzero = 0
        total_conv_nonzero = 0

        for N, mask, layer in zip(
            self.weights_in_layer, self.backward_masks, self.layers
        ):
            actual_S = torch.sum(layer.weight[mask == 0] == 0).item()
            N_str += "%i/%i, " % (N - actual_S, N)
            sp_p = float(N - actual_S) / float(N) * 100
            S_str += "%.2f%%, " % sp_p
            sparsity_percentages.append(sp_p)
            total_params += N
            total_nonzero += N - actual_S
            # if not is_linear:
            total_conv_nonzero += N - actual_S
            total_conv_params += N

        N_str = N_str[:-2] + "]"
        S_str = S_str[:-2] + "]"

        s += "nonzero_params=" + N_str + ",\n"
        s += "nonzero_percentages=" + S_str + ",\n"
        s += (
            "total_nonzero_params="
            + (
                "%i/%i (%.2f%%)"
                % (
                    total_nonzero,
                    total_params,
                    float(total_nonzero) / float(total_params) * 100,
                )
            )
            + ",\n"
        )
        s += (
            "total_CONV_nonzero_params="
            + (
                "%i/%i (%.2f%%)"
                % (
                    total_conv_nonzero,
                    total_conv_params,
                    float(total_conv_nonzero) / float(total_conv_params) * 100,
                )
            )
            + ",\n"
        )
        s += "step=" + str(self.step) + ",\n"
        s += "num_rigl_steps=" + str(self.rigl_steps) + ",\n"

        return s + ")"

    @torch.no_grad()
    def reset_momentum(self):
        # raise NotImplementedError("Not compatible with momentum yet")
        for layer, mask in zip(self.layers, self.backward_masks):

            param_state = self.optimizer.state[layer.weight]
            if "momentum_buffer" in param_state:
                # mask the momentum matrix
                buf = param_state["momentum_buffer"]
                buf *= mask

    @torch.no_grad()
    def apply_mask_to_weights(self):
        for layer, mask in zip(self.layers, self.backward_masks):
            layer.weight *= mask

    @torch.no_grad()
    def apply_mask_to_gradients(self):
        for layer, mask in zip(self.layers, self.backward_masks):
            # if sparsity is 0%, skip
            layer.weight.grad *= mask

    def check_if_backward_hook_should_accumulate_grad(self):
        """
        Used by the backward hooks. Basically just checks how far away the next rigl step is,
        if it's within `self.grad_accumulation_n` steps, return True.
        """

        if self.step >= self.T_end:
            return False

        steps_til_next_rigl_step = self.delta - (self.step % self.delta)
        return steps_til_next_rigl_step <= self.grad_accumulation_n

    def cosine_annealing(self):
        return self.alpha / 2 * (1 + np.cos((self.step * np.pi) / self.T_end))

    def __call__(self):
        self.step += 1
        if (self.step % self.delta) == 0 and self.step < self.T_end:  # check schedule
            self._rigl_step()
            self.rigl_steps += 1
            return False
        return True

    @torch.no_grad()
    def _rigl_step(self):
        drop_fraction = self.cosine_annealing()

        for l, layer in enumerate(self.layers):
            weight_sizes = layer.weight.shape
            if (
                int(self.num_connections * drop_fraction) <= 0
                or weight_sizes[1] * weight_sizes[2] * weight_sizes[3]
                <= self.num_connections
            ):
                continue

            channel_masks = []
            channel_new_weights = []
            channels = layer.weight.shape[0]
            for channel in range(channels):
                current_mask = self.backward_masks[l][channel]

                # calculate raw scores
                score_drop = torch.abs(layer.weight[channel])
                score_grow = torch.abs(
                    self.backward_hook_objects[l].dense_grad[channel]
                )

                # calculate drop/grow quantities
                n_total = self.weights_in_layer[l] // channels
                n_ones = torch.sum(current_mask).item()
                n_prune = int(n_ones * drop_fraction)
                n_keep = n_ones - n_prune

                # create drop mask
                _, sorted_indices = torch.topk(score_drop.view(-1), k=n_total)
                new_values = torch.where(
                    torch.arange(n_total, device=layer.weight.device) < n_keep,
                    torch.ones_like(sorted_indices),
                    torch.zeros_like(sorted_indices),
                )
                mask1 = new_values.scatter(0, sorted_indices, new_values)

                # flatten grow scores
                score_grow = score_grow.view(-1)

                # set scores of the enabled connections(ones) to min(s) - 1, so that they have the lowest scores
                score_grow_lifted = torch.where(
                    mask1 == 1,
                    torch.ones_like(mask1) * (torch.min(score_grow) - 1),
                    score_grow,
                )

                # create grow mask
                _, sorted_indices = torch.topk(score_grow_lifted, k=n_total)
                new_values = torch.where(
                    torch.arange(n_total, device=layer.weight.device) < n_prune,
                    torch.ones_like(sorted_indices),
                    torch.zeros_like(sorted_indices),
                )
                mask2 = new_values.scatter(0, sorted_indices, new_values)

                mask2_reshaped = torch.reshape(mask2, current_mask.shape)
                grow_tensor = torch.zeros_like(layer.weight[0])

                new_connections = (mask2_reshaped == 1) & (current_mask == 0)

                # update new weights to be initialized as zeros and update the weight tensors
                new_weights = torch.where(
                    new_connections.to(layer.weight.device),
                    grow_tensor,
                    layer.weight[channel],
                )
                channel_new_weights.append(new_weights)
                mask_combined = torch.reshape(mask1 + mask2, current_mask.shape).bool()
                channel_masks.append(mask_combined)

            layer.weight.data = torch.stack(channel_new_weights, 0)

            current_mask = self.backward_masks[l]
            # update the mask
            current_mask.data = torch.stack(channel_masks)

        self.reset_momentum()
        self.apply_mask_to_weights()
        self.apply_mask_to_gradients()



from abc import ABC, abstractmethod
from string import Template
from typing import Iterable, Iterator, Protocol, Sequence

import numpy as np

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import ToLogicEncoder
from elasticai.creator.vhdl.templates.utils import expand_multiline_template


# noinspection PyPropertyDefinition
class IOTable(Iterable[tuple[int, int]], Protocol):
    @property
    def input_bit_width(self) -> int:
        ...

    @property
    def output_bit_width(self) -> int:
        ...


class _EncodedIOTable(IOTable):
    def __iter__(self) -> Iterator[tuple[int, int]]:
        for i in range(len(self)):
            yield self[i]

    def __init__(
        self,
        input_encoder: ToLogicEncoder,
        output_encoder: ToLogicEncoder,
        inputs: Sequence[int],
        outputs: Sequence[int],
    ):
        self._input_encoder = input_encoder
        self._output_encoder = output_encoder
        self._inputs = inputs
        self._outputs = outputs

    def __getitem__(self, item) -> tuple[str, str]:
        encoded_input = self._input_encoder(self._inputs.__getitem__(item))
        encoded_output = self._output_encoder(self._outputs.__getitem__(item))
        return encoded_input, encoded_output

    def __len__(self):
        return min(len(self._inputs), len(self._outputs))

    @property
    def input_bit_width(self) -> int:
        return self._input_encoder.bit_width

    @property
    def output_bit_width(self) -> int:
        return self._output_encoder.bit_width


class TruthTableVHDLDesign(ABC):
    def __init__(self, io_table: IOTable, entity_name: str, architecture_name: str):
        self._table = io_table
        self.entity_name = entity_name
        self.architecture_name = architecture_name

    @abstractmethod
    def __call__(self, *args, **kwargs):
        ...

    @property
    def input_bitwidth(self) -> int:
        return self._table.input_bit_width

    @property
    def output_bitwidth(self) -> int:
        return self._table.output_bit_width

    def __iter__(self):
        yield from self.__call__()

    @classmethod
    def by_enumerating_output_sequence(
        cls, entity_name: str, architecture_name: str, outputs: Sequence[int]
    ):
        num_outputs = len(outputs)
        num_inputs = num_outputs
        inputs = range(num_inputs)

        input_encoder = ToLogicEncoder()
        output_encoder = ToLogicEncoder()
        output_encoder.register_symbols(np.unique(outputs))
        input_encoder.register_symbols(range(num_inputs))

        io_table = _EncodedIOTable(input_encoder, output_encoder, inputs, outputs)
        vhdl_lut_design = cls(io_table, entity_name, architecture_name)
        return vhdl_lut_design


class TruthTableVHDLDesignCaseWhen(TruthTableVHDLDesign):
    def __call__(self) -> Code:
        cases = [
            f'when "{input}" => output <= "{output}";' for input, output in self._table
        ]
        code = expand_multiline_template(
            read_text(
                "elasticai.creator.vhdl.templates.precomputed_convs",
                "truth_table.tpl.vhd",
            ),
            cases=cases,
        )

        def substitution(line: str):
            return Template(line).substitute(
                entity_name=self.entity_name,
                architecture_name=self.architecture_name,
                input_vector_start_bit=self._table.input_bit_width - 1,
                output_vector_start_bit=self._table.output_bit_width - 1,
            )

        yield from map(substitution, code)



import os
from argparse import ArgumentParser
from functools import partial

import numpy as np
import torch

from elasticai.creator.qat.layers import QLSTMCell
from elasticai.creator.resource_utils import copy_file
from elasticai.creator.vhdl.generator_functions_for_lstm import (
    extract_fixed_point_weights_and_bias,
    generate_rom_file,
    inference_model_on_random_data,
)
from elasticai.creator.vhdl.lstm_testbench_generator import LSTMCellTestBench
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import Sigmoid, Tanh
from elasticai.creator.vhdl.vhdl_formatter.vhdl_formatter import format_vhdl

"""
this module generates all vhd files for a single lstm cell
"""


def define_lstm_cell(input_size: int, hidden_size: int) -> QLSTMCell:
    """
    returns a QLSTM Cell with the given input and hidden size
    Args:
        input_size (int): input size of QLSTM Cell
        hidden_size (int): hidden size of QLSTM Cell
    Returns:
        returns the corresponding QLSTM Cell
    """
    return QLSTMCell(
        input_size=input_size,
        hidden_size=hidden_size,
        state_quantizer=lambda x: x,
        weight_quantizer=lambda x: x,
    )


def main() -> None:
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--path",
        help="path to folder for generated vhd files",
        required=True,
    )
    args = arg_parser.parse_args()
    if not os.path.isdir(args.path):
        os.mkdir(args.path)

    def destination_path(file_name: str) -> str:
        return os.path.join(args.path, file_name)

    # set the current values
    torch.manual_seed(0)
    frac_bits = 8
    total_bits = 16
    input_size = 1
    hidden_size = 20

    to_fp = partial(
        float_values_to_fixed_point, total_bits=total_bits, frac_bits=frac_bits
    )

    lstm_cell = define_lstm_cell(input_size, hidden_size)

    weights_list, bias_list = extract_fixed_point_weights_and_bias(
        lstm_cell, total_bits, frac_bits
    )
    print("weights_list:", [list(map(lambda x: x.to_hex(), w)) for w in weights_list])
    print("bias_list:", [list(map(lambda x: x.to_hex(), b)) for b in bias_list])

    x_h_test_input, c_test_input, h_output = inference_model_on_random_data(
        lstm_cell, total_bits, frac_bits
    )
    print("x_h_test_input:", list(map(lambda x: x.to_hex(), x_h_test_input)))
    print("c_test_input:", list(map(lambda x: x.to_hex(), c_test_input)))
    print("h_output:", list(map(lambda x: x.to_hex(), h_output)))

    # generate source files for use-case

    # generate weight and bias source files
    weight_names = ["wi", "wf", "wg", "wo", "bi", "bf", "bg", "bo"]
    for name, values in zip(weight_names, weights_list + bias_list):
        generate_rom_file(
            file_path=destination_path(f"{name}_rom.vhd"),
            rom_name=f"{name}_rom",
            values=values,
            resource_option="auto",
        )

    # generate sigmoid and tanh activation source files
    file_path_sigmoid = destination_path("sigmoid.vhd")
    with open(file_path_sigmoid, "w") as writer:
        sigmoid = Sigmoid(
            x=to_fp(np.linspace(-2.5, 2.5, 256).tolist()),
        )
        sigmoid_code = sigmoid()
        for line in sigmoid_code:
            writer.write(line + "\n")

    file_path_tanh = destination_path("tanh.vhd")
    with open(file_path_tanh, "w") as writer:
        tanh = Tanh(
            x=to_fp(np.linspace(-1, 1, 256).tolist()),
        )
        tanh_code = tanh()
        for line in tanh_code:
            writer.write(line + "\n")

    # generate testbench file for use-case
    file_path_testbench = destination_path("lstm_cell_tb.vhd")
    with open(file_path_testbench, "w") as writer:
        lstm_cell_tb = LSTMCellTestBench(
            input_size=input_size,
            hidden_size=hidden_size,
            test_x_h_data=x_h_test_input,
            test_c_data=c_test_input,
            h_out=h_output,
            component_name="lstm_cell",
        )
        lstm_cell_code = lstm_cell_tb()
        for line in lstm_cell_code:
            writer.write(line + "\n")

    for name in weight_names:
        format_vhdl(file_path=destination_path(f"{name}_rom.vhd"))

    format_vhdl(file_path=file_path_sigmoid)
    format_vhdl(file_path=file_path_tanh)
    format_vhdl(file_path=file_path_testbench)

    # copy static files
    for file in ["dual_port_2_clock_ram.vhd", "lstm_cell.vhd", "lstm_common.vhd"]:
        copy_file("elasticai.creator.vhdl.templates", file, destination_path(file))


if __name__ == "__main__":
    main()



import unittest


class TruthTableTestCase(unittest.TestCase):
    """
    Tests:
        - create output by_enumerating_output_sequence
    """

    @unittest.SkipTest
    def test_first(self):
        self.fail()



from torch import nn
from torch.nn import Conv1d, Conv2d, Linear
from torch.nn.utils import parametrize

"""
Modules that work as a sequence of 3  or more layers. Useful for writing more compact models
"""


# when applying constraints to blocks loop with model.modules()
# finish with nn.Identity as an activation if not using Softmax
class Conv1dBlock(nn.Module):
    """
    Sequence QConv1d - batchNorm - activation
    uses default batchNorm parameters. Most other parameters affect QConv1d
    @param conv_quantizer: The quantizer of the QConv1d
    @param activation: an instance of the activation
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        conv_quantizer,
        activation,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        constraints: list = None,
    ):
        super().__init__()

        self.conv1d = Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(self.conv1d, "weight", conv_quantizer)
        self.batch_norm = nn.BatchNorm1d(out_channels)
        self.activation = activation

    @property
    def codomain(self):
        if self.activation is not None and hasattr(self.activation, "codomain"):
            return self.activation.codomain
        return None

    def forward(self, input):
        x = self.conv1d(input)
        x = self.batch_norm(x)
        x = self.activation(x)
        return x


class Conv2dBlock(nn.Module):
    """
    Sequence Conv2d - batchNorm - activation
    uses default batchNorm parameters. Most other parameters affect QConv1d
    @param conv_quantizer: The quantizer of the QConv1d
    @param activation: an instance of the activation
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        conv_quantizer,
        activation,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
    ):
        super().__init__()

        self.conv2d = Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(self.conv2d, "weight", conv_quantizer)
        self.batch_norm = nn.BatchNorm2d(out_channels)
        self.activation = activation

    @property
    def codomain(self):
        if self.activation is not None and hasattr(self.activation, "codomain"):
            return self.activation.codomain
        return None

    def forward(self, input):
        x = self.conv2d(input)
        x = self.batch_norm(x)
        x = self.activation(x)
        return x


class LinearBlock(nn.Module):
    """
    Sequence Linear - batchNorm - activation
    uses default batchNorm parameters. Most other parameters affect Qconv1d
    @param linear_quantizer: The quantizer of the linear weight
    @param activation: an instance of the activation
    """

    def __init__(
        self,
        in_features,
        out_features,
        linear_quantizer,
        activation,
        bias=False,
        constraints: list = None,
    ):
        super().__init__()
        self.linear = Linear(
            in_features,
            out_features,
            bias=bias,
        )
        if linear_quantizer is not None:
            parametrize.register_parametrization(
                self.linear, "weight", linear_quantizer
            )
        self.batchnorm = nn.BatchNorm1d(out_features)
        self.activation = activation

    def forward(self, input):
        x = self.linear(input)
        x = self.batchnorm(x)
        x = self.activation(x)
        return x


class DepthwiseConv1dBlock(nn.Module):
    """
    Sequence depthwise Conv1d - batchNorm -activation, 1x1 Conv1d - batchnorm - activation
    uses default batchNorm parameters. Most other parameters affect Qconv1d
    Args:
     conv_quantizer: The quantizer of the first QConv1d
     activation: an instance of the activation  after the first batch norm,
     pointwise_quantizer: the quantizer of the second Qconv1d
     pointwise_activation: an instance of the activation after the second batch norm,
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        conv_quantizer,
        pointwise_quantizer,
        activation,
        pointwise_activation,
        stride=1,
        padding=0,
        dilation=1,
        bias=True,
        padding_mode="zeros",
        constraints: list = None,
        pointwise_constraints: list = None,
    ):
        super().__init__()

        self.depthwiseConv1d = Conv1d(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(
                self.depthwiseConv1d, "weight", conv_quantizer
            )
        self.batchnorm = nn.BatchNorm1d(in_channels)
        self.activation = activation

        self.pointwiseConv1d = Conv1d(
            in_channels,
            out_channels,
            1,
            stride=1,
            groups=1,
            bias=bias,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(
                self.pointwiseConv1d, "weight", pointwise_quantizer
            )
        self.pointwise_batchnorm = nn.BatchNorm1d(out_channels)
        self.pointwise_activation = pointwise_activation

    @property
    def codomain(self):
        if self.pointwise_activation is not None and hasattr(
            self.pointwise_activation, "codomain"
        ):
            return self.pointwise_activation.codomain
        return None

    def forward(self, input):
        x = self.depthwiseConv1d(input)
        x = self.batchnorm(x)
        x = self.activation(x)
        x = self.pointwiseConv1d(x)
        x = self.pointwise_batchnorm(x)
        x = self.pointwise_activation(x)
        return x



"""
The module contains classes and functions for generating vhdl code similar to the language module
This module includes CodeGenerator that are only used by the vhdl testbenches
"""
import abc
from abc import abstractmethod
from typing import Iterator


class TestBenchBase(abc.ABC):
    simulation_start_msg = 'report "======Simulation Start======" severity Note'
    simulation_end_msgs = (
        'report "======Simulation Success======" severity Note',
        'report "Please check the output message." severity Note',
        "wait",
    )

    @abstractmethod
    def _body(self) -> Iterator[str]:
        ...

    def __iter__(self) -> Iterator[str]:
        yield self.simulation_start_msg
        yield from self._body()
        yield from self.simulation_end_msgs



from functools import partial

import numpy as np
import torch
from torch.nn import LSTM, LSTMCell

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
)
from elasticai.creator.vhdl.rom import Rom


def generate_rom_file(
    file_path: str,
    rom_name: str,
    values: list[FixedPoint],
    resource_option: str,
) -> None:
    """
    generates the rom files for a given list of values
    Args:
        file_path (str): paths where files should be stored
        values (list[list[int]]): list with four lists with the fixed point values for each weight or bias
        rom_name (str): name for the file
        resource_option (str): resource option
    """
    with open(file_path, "w") as writer:
        rom = Rom(
            rom_name=rom_name,
            values=values,
            resource_option=resource_option,
        )
        rom_code = rom()
        for line in rom_code:
            writer.write(line + "\n")


def inference_model_on_random_data(
    lstm: LSTMCell | LSTM,
    total_bits: int,
    frac_bits: int,
) -> tuple[list[FixedPoint], list[FixedPoint], list[FixedPoint]]:
    """
    do inference on defined LSTM Cell or LSTM
    Args:
        lstm (LSTMCell | LSTM): current LSTM Cell or LSTM
        total_bits (int): total number of bits for one fixed point number
        frac_bits (int): number of fraction bits for one fixed point number
    Returns:
        returns three lists/arrays
        the first and second list are the x_h input_data and cx of the lstm cell
        the third array is the hx of the lstm cell
    """
    torch.manual_seed(0)

    input_data = torch.randn(2, 1, lstm.input_size)  # (time_steps, batch, input_size)
    hx = torch.randn(1, lstm.hidden_size)  # this is the hidden states
    cx = torch.randn(1, lstm.hidden_size)  # this the cell states

    x_h_input = np.hstack(([], [], []))

    for i in range(input_data.size()[0]):
        x_h_input = np.hstack(
            (input_data[i].detach().numpy().flatten(), hx.detach().numpy().flatten())
        )
        result = lstm(input_data[i], (hx, cx))

        if isinstance(lstm, LSTMCell):
            hx, cx = result
        else:
            _, (hx, cx) = result

    to_fp = partial(
        float_values_to_fixed_point, total_bits=total_bits, frac_bits=frac_bits
    )

    return (
        to_fp(x_h_input),
        to_fp(cx.detach().numpy().flatten()),
        to_fp(hx.detach().numpy().flatten()),
    )


def extract_fixed_point_weights_and_bias(
    lstm: LSTMCell | LSTM, total_bits: int, frac_bits: int
) -> tuple[list[list[FixedPoint]], list[list[FixedPoint]]]:
    """
    calculates the weights and bias for the 1st layer of the given multilayer LSTM
    or the given LSTMCell
    Args:
        lstm (LSTMCell | LSTM): current LSTM object
        total_bits (int): total number of bits for one fixed point number
        frac_bits (int): number of fraction bits for one fixed point number
    Returns:
        returns two lists, one for the weights and one for the bias
        in each list are four list of FixedPoint numbers of the weights or bias
    """
    if isinstance(lstm, LSTMCell):
        weight_ih = lstm.weight_ih.detach().numpy()
        weight_hh = lstm.weight_hh.detach().numpy()
        bias_ih = lstm.bias_ih.detach().numpy()
        bias_hh = lstm.bias_hh.detach().numpy()
    else:
        weight_ih = lstm.weight_ih_l0.detach().numpy()
        weight_hh = lstm.weight_hh_l0.detach().numpy()
        bias_ih = lstm.bias_ih_l0.detach().numpy()
        bias_hh = lstm.bias_hh_l0.detach().numpy()

    weights = np.hstack((weight_ih, weight_hh)).flatten().flatten()
    bias = bias_hh + bias_ih

    len_weights = (lstm.input_size + lstm.hidden_size) * lstm.hidden_size
    len_bias = lstm.hidden_size

    wi = weights[len_weights * 0 : len_weights * 1]  # [Wii, Whi]
    wf = weights[len_weights * 1 : len_weights * 2]  # [Wif, Whf]
    wg = weights[len_weights * 2 : len_weights * 3]  # [Wig, Whg]
    wo = weights[len_weights * 3 : len_weights * 4]  # [Wio, Who]

    bi = bias[len_bias * 0 : len_bias * 1]  # B_ii+B_hi
    bf = bias[len_bias * 1 : len_bias * 2]  # B_if+B_hf
    bg = bias[len_bias * 2 : len_bias * 3]  # B_ig+B_hg
    bo = bias[len_bias * 3 : len_bias * 4]  # B_io+B_ho

    to_fp = partial(
        float_values_to_fixed_point, total_bits=total_bits, frac_bits=frac_bits
    )

    fixed_point_weights = list(map(to_fp, [wi, wf, wg, wo]))
    fixed_point_bias = list(map(to_fp, [bi, bf, bg, bo]))

    return fixed_point_weights, fixed_point_bias



import subprocess

from elasticai.creator.resource_utils import get_full_path

"""
    we are using vsg (VHDL Style Guide) for indenting the generated vhdl files
    config.json defines the indentation size, since the default indentation size is 2 spaces

    source :
        https://github.com/jeremiah-c-leary/vhdl-style-guide/tree/master
"""


def format_vhdl(file_path: str):
    config_path = get_full_path("elasticai.creator.vhdl.vhdl_formatter", "config.json")
    subprocess.run(
        f"vsg -f {file_path} --style indent_only --configuration {config_path} --fix",
        shell=True,
    )



from elasticai.creator.tests.vhdl.vhdl_file_testcase import GeneratedVHDLCodeTest
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.rom import Rom


class GenerateROMVhdTest(GeneratedVHDLCodeTest):
    def test_compare_files(self) -> None:
        # biases for the input gate
        bi = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

        generate_rom = Rom(
            rom_name="rom_bi",
            values=float_values_to_fixed_point(bi, total_bits=12, frac_bits=4),
            resource_option="auto",
        )
        generated_code = list(generate_rom())

        expected_code = [
            "library ieee;",
            "use ieee.std_logic_1164.all;",
            "use ieee.std_logic_unsigned.all;",
            "entity rom_bi is",
            "port (",
            "clk : in std_logic;",
            "en : in std_logic;",
            "addr : in std_logic_vector(3-1 downto 0);",
            "data : out std_logic_vector(12-1 downto 0)",
            ");",
            "end entity rom_bi;",
            "architecture rtl of rom_bi is",
            "type rom_bi_array_t is array (0 to 2**3-1) of std_logic_vector(12-1 downto 0);",
            'signal ROM : rom_bi_array_t:=(x"011",x"023",x"034",x"046",x"058",x"069",x"000",x"000");',
            "attribute rom_style : string;",
            'attribute rom_style of ROM : signal is "auto";',
            "begin",
            "ROM_process: process(clk)",
            "begin",
            "if rising_edge(clk) then",
            "if (en = '1') then",
            "data <= ROM(conv_integer(addr));",
            "end if;",
            "end if;",
            "end process ROM_process;",
            "end architecture rtl;",
        ]

        self.assertEqual(expected_code, generated_code)



"""Implementation of the breakdown. Used to represent multiple layers with
a fixed width.
"""
import torch.nn
from torch.nn.utils import parametrize

from elasticai.creator.qat.blocks import Conv2dBlock
from elasticai.creator.qat.layers import ChannelShuffle
from elasticai.creator.qat.masks import FixedOffsetMask4d


def generate_conv2d_sequence_with_width(
    in_channels: int,
    out_channels: int,
    channel_width: int,
    kernel_size,
    weight_quantization,
    activation,
):
    """
    This will generate a sequential model composed of multiple layers, each a
    weight with channel width length.
    After the first block the channels are shuffled so the information of each
    output channel is composed from one in each input channel.
    eg: 256x256 1x1 convolution with channel width of 8. Each channel of the
     first layer will join information of 8 points  making 256/8 = 32 parts per out channel.
    So 32 groups and 256(out) = 256*32 output channels in the first layer. the
     second layer will again join 8  of those 32 resulting in 256*4
    groups and out channels. The last to keep the width will have a smaller
     number of groups e.g 256*4/8 = 128 groups. This has the effect of making very small kernels,
    being able to be represented by a much smaller amount of LUTs than a 256:1
     mapping.

    Args:
        in_channels:
        out_channels:
        channel_width:
        kernel_size:
        weight_quantization:
        activation:

    Returns:

    """
    layers: list[Conv2dBlock | ChannelShuffle] = []
    if in_channels < channel_width:
        raise ValueError(
            "Channel width cannot be bigger than the number of input channels"
        )
    next_in_channels = in_channels
    next_groups = in_channels // channel_width
    next_out_channels = out_channels * next_groups

    while next_out_channels >= out_channels:
        layers.append(
            Conv2dBlock(
                in_channels=next_in_channels,
                out_channels=next_out_channels,
                kernel_size=kernel_size,
                activation=activation,
                conv_quantizer=weight_quantization,
                groups=next_groups,
            )
        )
        if len(layers) == 1:
            layers.append(ChannelShuffle(groups=next_groups))
        if next_out_channels == out_channels:
            break
        next_in_channels = next_out_channels
        next_groups = (
            next_groups // channel_width
            if len(layers) > 2
            else next_out_channels // channel_width
        )
        next_out_channels = max(next_groups, out_channels)

    return torch.nn.Sequential(*layers)


class BreakdownConv2dBlock(torch.nn.Module):
    """
    Implementation of the kernel breakdown where each each channel will only
     select one row of the original nxn offset by
    a module of its own index.
    Args:
     conv_quantizer: The quantizer of the first QConv1d
     activation: an instance of the activation  after the first batch norm,
     pointwise_quantizer: the quantizer of the second Qconv1d
     pointwise_activation: an instance of the activation after the second
                           batch norm,
     pointwise_channel_width: the channel width of the pointwise convolution
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        groups,
        activation,
        pointwise_activation,
        conv_quantizer=None,
        pointwise_quantizer=None,
        stride=1,
        padding=0,
        dilation=1,
        bias=True,
        padding_mode="zeros",
    ):

        if isinstance(kernel_size, tuple):
            breakdown_multiplier = kernel_size[0] * groups
        else:
            breakdown_multiplier = kernel_size * groups
        super().__init__()
        self.Conv2d = Conv2dBlock(
            in_channels=in_channels,
            out_channels=out_channels * breakdown_multiplier,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
            activation=activation,
            conv_quantizer=conv_quantizer,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(
                self.depthwiseConv2d, "weight", conv_quantizer
            )
        mask = FixedOffsetMask4d(
            out_channels=out_channels * breakdown_multiplier,
            in_channels=in_channels,
            kernel_size=kernel_size,
            offset_axis=2,
            axis_width=1,
            groups=groups,
        )
        parametrize.register_parametrization(self.Conv2d.conv2d, "weight", mask)
        self.shuffle = ChannelShuffle(groups=groups)
        self.pointwiseConv2d = Conv2dBlock(
            in_channels=out_channels * breakdown_multiplier,
            out_channels=out_channels,
            kernel_size=1,
            stride=stride,
            dilation=dilation,
            groups=out_channels,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
            activation=pointwise_activation,
            conv_quantizer=pointwise_quantizer,
        )

    @property
    def codomain(self):
        if self.pointwise_activation is not None and hasattr(
            self.pointwise_activation, "codomain"
        ):
            return self.pointwise_activation.codomain
        return None

    def forward(self, input):
        x = self.Conv2d(input)
        x = self.shuffle(x)
        x = self.pointwiseConv2d(x)
        return x


class depthwisePointwiseBreakdownConv2dBlock(torch.nn.Module):
    """
    Sequence depthwise QConv2d - batchNorm -activation, pointwise_breakdown
    uses default batchNorm parameters. Most parameters affect the depthwise Qconv2d,
    Args:
     conv_quantizer: The weightr quantizer of the first Conv2d can be none
     activation: an instance of the activation  after the first batch norm,
     pointwise_quantizer: the quantizer of the pointwise Conv2d can be none
     pointwise_activation: an instance of the activation after each pointwise batch norm,
     pointwise_channel_width: the channel width of the pointwise convolution
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        activation,
        pointwise_activation,
        pointwise_channel_width,
        conv_quantizer=None,
        pointwise_quantizer=None,
        stride=1,
        padding=0,
        dilation=1,
        bias=True,
        padding_mode="zeros",
    ):
        super().__init__()
        self.depthwiseConv2d = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
            padding=padding,
            padding_mode=padding_mode,
        )
        if conv_quantizer is not None:
            parametrize.register_parametrization(
                self.depthwiseConv2d, "weight", conv_quantizer
            )
        self.batchnorm = torch.nn.BatchNorm2d(in_channels)
        self.activation = activation

        self.pointwiseConv2d = generate_conv2d_sequence_with_width(
            in_channels,
            out_channels,
            pointwise_channel_width,
            1,
            pointwise_quantizer,
            pointwise_activation,
        )

    @property
    def codomain(self):
        if self.pointwise_activation is not None and hasattr(
            self.pointwise_activation, "codomain"
        ):
            return self.pointwise_activation.codomain
        return None

    def forward(self, input):
        x = self.depthwiseConv2d(input)
        x = self.batchnorm(x)
        x = self.activation(x)
        x = self.pointwiseConv2d(x)
        return x



import functools
from typing import Any, Callable

"""
The meta programming functions in this module are mostly used to implement the decorator software design patterns
that allow users to add support for elastic ai functionalities, such as *precomputing* to custom `torch.nn.Module`s.
"""


def add_method(method_name: str, method: Callable) -> Callable[[type], type]:
    """Python decorator for adding a method to a class definition.

    Returns:
        A class with a method called `method_name` added, that is implemented by `method`.

    """

    def wrapper(cls: type) -> type:
        return cls

    return wrapper


def add_instance_attribute(
    attribute_name: str, attribute: Any
) -> Callable[[type], type]:
    """Python decorator that adds an instance attribute to a class definition. This behaves as if the assignment was
    `self.attribute_name = attribute` was issued after the call to the original `__init__` function of the class.
    Returns:
        A new class being identical to the original class definition, except for the `__init__` function that executes above assignment
        as a last step

    """

    def adds(cls: type) -> type:
        init: Any = cls.__init__

        @functools.wraps(init)
        def wrapped_init(self, *args, **kwargs):
            init(self, *args, **kwargs)
            setattr(self, attribute_name, attribute)

        setattr(cls, "__init__", wrapped_init)
        return cls

    return adds



import os
from argparse import ArgumentParser

import numpy as np

from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.precomputed_scalar_function import Tanh
from elasticai.creator.vhdl.vhdl_formatter.vhdl_formatter import format_vhdl


def main():
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--path",
        help="path to folder for generated vhd files",
        required=True,
    )
    args = arg_parser.parse_args()
    if not os.path.isdir(args.path):
        os.mkdir(args.path)
    component_name = "tanh"
    file_path = os.path.join(args.path, f"{component_name}.vhd")

    # noinspection PyTypeChecker
    data = float_values_to_fixed_point(
        np.linspace(-5, 5, 259).tolist(), total_bits=16, frac_bits=8
    )

    tanh = Tanh(
        component_name=component_name,
        x=data,
    )
    tanh_code = tanh()

    with open(file_path, "w") as writer:
        for line in tanh_code:
            writer.write(line + "\n")

    format_vhdl(file_path=file_path)


if __name__ == "__main__":
    main()



class WeightClipper:
    """
    This is an applied constraint meant to be used  after each batch. It clips the weights of the given module to be
     between -1,1.
    Meant to be used with binarization
    """

    def __init__(self):
        pass

    def __call__(self, module):
        module.weight.clip_(-1, 1)



import math
from itertools import chain
from typing import Iterable, Iterator

from elasticai.creator.vhdl.language import (
    Architecture,
    Code,
    ContextClause,
    Entity,
    Keywords,
    LibraryClause,
    PortMap,
    Procedure,
    Process,
    UseClause,
    hex_representation,
)
from elasticai.creator.vhdl.language_testbench import TestBenchBase
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    infer_total_and_frac_bits,
)


class LSTMCellTestBench:
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        test_x_h_data: list[FixedPoint],
        test_c_data: list[FixedPoint],
        h_out: list[FixedPoint],
        component_name: str = None,
    ):
        self.component_name = self._get_lower_case_class_name_or_component_name(
            component_name=component_name
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.x_h_addr_width = math.ceil(math.log2(input_size + hidden_size))
        self.hidden_addr_width = max(1, math.ceil(math.log2(hidden_size)))
        self.w_addr_width = math.ceil(
            math.log2((input_size + hidden_size) * hidden_size)
        )
        self.data_width, self.frac_width = infer_total_and_frac_bits(
            test_x_h_data, test_c_data, h_out
        )

        def to_hex(value: FixedPoint) -> str:
            return hex_representation(value.to_hex())

        self.test_x_h_data = list(map(to_hex, test_x_h_data))
        self.test_c_data = list(map(to_hex, test_c_data))
        self.h_out = h_out

    @classmethod
    def _get_lower_case_class_name_or_component_name(cls, component_name):
        if component_name is None:
            return cls.__name__.lower()
        return component_name

    @property
    def file_name(self) -> str:
        return f"{self.component_name}_tb.vhd"

    def __call__(self) -> Iterable[str]:
        library = ContextClause(
            library_clause=LibraryClause(logical_name_list=["ieee", "work"]),
            use_clause=UseClause(
                selected_names=[
                    "ieee.std_logic_1164.all",
                    "ieee.numeric_std.all",
                    "work.lstm_common.all",
                ]
            ),
        )

        entity = Entity(self.component_name + "_tb")
        entity.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
            f"INPUT_SIZE : integer := {self.input_size}",
            f"HIDDEN_SIZE : integer := {self.hidden_size}",
            f"X_H_ADDR_WIDTH : integer := {self.x_h_addr_width}",
            f"HIDDEN_ADDR_WIDTH : integer := {self.hidden_addr_width}",
            f"W_ADDR_WIDTH : integer := {self.w_addr_width}",
        ]
        entity.port_list = [
            "clk : out std_logic",
        ]

        procedure_0 = Procedure(identifier="send_x_h_data")
        procedure_0.declaration_list = [
            "addr_in : in std_logic_vector(X_H_ADDR_WIDTH-1 downto 0)",
            "data_in : in std_logic_vector(DATA_WIDTH-1 downto 0)",
            "signal clock : in std_logic",
            "signal wr : out std_logic",
            "signal addr_out : out std_logic_vector(X_H_ADDR_WIDTH-1 downto 0)",
        ]
        procedure_0.declaration_list_with_is = [
            "signal data_out : out std_logic_vector(DATA_WIDTH-1 downto 0))",
        ]
        procedure_0.statement_list = [
            "addr_out <= addr_in",
            "data_out <= data_in",
            "wait until clock='0'",
            "wr <= '1'",
            "wait until clock='1'",
            "wait until clock='0'",
            "wr <= '0'",
            "wait until clock='1'",
        ]
        procedure_1 = Procedure(identifier="send_c_data")
        procedure_1.declaration_list = [
            "addr_in : in std_logic_vector(HIDDEN_ADDR_WIDTH-1 downto 0)",
            "data_in : in std_logic_vector(DATA_WIDTH-1 downto 0)",
            "signal clock : in std_logic",
            "signal wr : out std_logic",
            "signal addr_out : out std_logic_vector(HIDDEN_ADDR_WIDTH-1 downto 0)",
        ]
        procedure_1.declaration_list_with_is = [
            "signal data_out : out std_logic_vector(DATA_WIDTH-1 downto 0))",
        ]
        procedure_1.statement_list = [
            "addr_out <= addr_in",
            "data_out <= data_in",
            "wait until clock='0'",
            "wr <= '1'",
            "wait until clock='1'",
            "wait until clock='0'",
            "wr <= '0'",
            "wait until clock='1'",
        ]

        process = Process(
            identifier="clock",
        )
        process.process_statements_list = [
            "clock <= '0'",
            "wait for clk_period/2",
            "clock <= '1'",
            "wait for clk_period/2",
        ]

        uut_port_map = PortMap(
            map_name="uut",
            component_name="entity work." + self.component_name + "(rtl)",
        )
        uut_port_map.generic_map_list = (
            "DATA_WIDTH => DATA_WIDTH",
            "FRAC_WIDTH => FRAC_WIDTH",
            "INPUT_SIZE => INPUT_SIZE",
            "HIDDEN_SIZE => HIDDEN_SIZE",
            "X_H_ADDR_WIDTH => X_H_ADDR_WIDTH",
            "HIDDEN_ADDR_WIDTH => HIDDEN_ADDR_WIDTH",
            "W_ADDR_WIDTH => W_ADDR_WIDTH",
        )
        uut_port_map.signal_list.append("clock => clock")
        uut_port_map.signal_list.append("reset => reset")
        uut_port_map.signal_list.append("enable => enable")
        uut_port_map.signal_list.append("x_h_we => x_config_en")
        uut_port_map.signal_list.append("x_h_data => x_config_data")
        uut_port_map.signal_list.append("x_h_addr => x_config_addr")
        uut_port_map.signal_list.append("c_we => c_config_en")
        uut_port_map.signal_list.append("c_data_in => c_config_data")
        uut_port_map.signal_list.append("c_addr_in => c_config_addr")
        uut_port_map.signal_list.append("done => done")
        uut_port_map.signal_list.append("h_out_en => h_out_en")
        uut_port_map.signal_list.append("h_out_data => h_out_data")
        uut_port_map.signal_list.append("h_out_addr => h_out_addr")

        test_cases = TestCasesLSTMCell(
            reference_h_out=self.h_out,
            input_size=self.input_size,
            hidden_size=self.hidden_size,
        )
        test_process = Process(identifier="test")
        test_process.process_statements_list = [t for t in test_cases()]

        architecture = Architecture(
            design_unit=self.component_name + "_tb",
        )
        architecture.architecture_declaration_list.append(
            "signal clk_period : time := 10 ns"
        )

        architecture.architecture_declaration_list.append("signal clock : std_logic")
        architecture.architecture_declaration_list.append("signal enable : std_logic")
        architecture.architecture_declaration_list.append(
            "signal reset: std_logic:='0'"
        )
        architecture.architecture_declaration_list.append(
            "signal x_config_en: std_logic:='0'"
        )
        architecture.architecture_declaration_list.append(
            "signal x_config_data:std_logic_vector(DATA_WIDTH-1 downto 0):=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal x_config_addr:std_logic_vector(X_H_ADDR_WIDTH-1 downto 0) :=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal h_config_en: std_logic:='0'"
        )
        architecture.architecture_declaration_list.append(
            "signal h_config_data:std_logic_vector(DATA_WIDTH-1 downto 0):=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal h_config_addr:std_logic_vector(HIDDEN_ADDR_WIDTH-1 downto 0) :=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal c_config_en: std_logic:='0'"
        )
        architecture.architecture_declaration_list.append(
            "signal c_config_data:std_logic_vector(DATA_WIDTH-1 downto 0):=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal c_config_addr:std_logic_vector(HIDDEN_ADDR_WIDTH-1 downto 0) :=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal done :  std_logic:='0'"
        )
        architecture.architecture_declaration_list.append("signal h_out_en : std_logic")
        architecture.architecture_declaration_list.append(
            "signal h_out_addr : std_logic_vector(HIDDEN_ADDR_WIDTH-1 downto 0) :=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "signal h_out_data : std_logic_vector(DATA_WIDTH-1 downto 0):=(others=>'0')"
        )
        architecture.architecture_declaration_list.append(
            "type X_H_ARRAY is array (0 to 511) of signed(16-1 downto 0)"
        )
        architecture.architecture_declaration_list.append(
            "type C_ARRAY is array (0 to 511) of signed(16-1 downto 0)"
        )
        architecture.architecture_declaration_list.append(
            f"signal test_x_h_data : X_H_ARRAY := ({','.join(self.test_x_h_data)},others=>(others=>'0'))"
        )
        architecture.architecture_declaration_list.append(
            f"signal test_c_data : C_ARRAY := ({','.join(self.test_c_data)},others=>(others=>'0'))"
        )
        architecture.architecture_component_list.append(procedure_0)
        architecture.architecture_component_list.append(procedure_1)
        architecture.architecture_process_list.append(process)
        architecture.architecture_port_map_list.append(uut_port_map)
        architecture.architecture_assignment_at_end_of_declaration_list.append(
            "clk <= clock"
        )
        architecture.architecture_statement_part = test_process

        code = chain(library(), entity(), architecture())
        return code


class TestCasesLSTMCommonGate(TestBenchBase):
    def __init__(
        self,
        x_mem_list_for_testing: list[FixedPoint],
        w_mem_list_for_testing: list[FixedPoint],
        b_list_for_testing: list[FixedPoint],
        y_list_for_testing: list[FixedPoint],
        y_variable_name: str = "y",
    ):
        assert (
            len(x_mem_list_for_testing)
            == len(w_mem_list_for_testing)
            == len(b_list_for_testing)
            == len(y_list_for_testing)
        )

        def to_hex(value: FixedPoint) -> str:
            return hex_representation(value.to_hex())

        self.x_mem_list_for_testing = list(map(to_hex, x_mem_list_for_testing))
        self.w_mem_list_for_testing = list(map(to_hex, w_mem_list_for_testing))
        self.b_list_for_testing = list(map(to_hex, b_list_for_testing))
        self.y_list_for_testing = list(map(int, y_list_for_testing))
        self.y_variable_name = y_variable_name

    def _body(self) -> Iterator[str]:
        counter = 0
        yield f"vector_len <= to_unsigned(10, VECTOR_LEN_WIDTH)"
        for x_mem_value, w_mem_value, b, y_value in zip(
            self.x_mem_list_for_testing,
            self.w_mem_list_for_testing,
            self.b_list_for_testing,
            self.y_list_for_testing,
        ):
            yield f"X_MEM <= ({x_mem_value})"
            yield f"W_MEM <= ({w_mem_value})"
            yield f"b <= {b}"
            yield from (
                "reset <= '1'",
                "wait for 2*clk_period",
                "wait until clock = '0'",
                "reset <= '0'",
                "wait until ready = '1'",
            )

            yield f"report \"expected output is {y_value}, value of '{self.y_variable_name}' is \" & integer'image(to_integer(signed({self.y_variable_name})))"
            yield f'assert {self.y_variable_name}={y_value} report "The {counter}. test case fail" severity error'
            yield "reset <= '1'"
            yield "wait for 1*clk_period"
            counter = counter + 1

    def __call__(self):
        yield from iter(self)


class TestCasesLSTMCell(TestBenchBase):
    def __init__(
        self,
        reference_h_out: list[FixedPoint],
        input_size: int = 0,
        hidden_size: int = 0,
    ):
        self.reference_h_out = list(map(int, reference_h_out))

        assert (input_size != 0) and (
            hidden_size != 0
        ), "hidden_size and input_size is not set yet"

        self.len_of_x_h_vector = input_size + hidden_size
        self.len_of_cell_vector = hidden_size
        self.len_of_h_vector = hidden_size

    def _body(self) -> Iterator[str]:
        yield f"reset <= '1'"
        yield f"h_out_en <= '0'"
        yield f"wait for 2*clk_period"
        yield f"reset <= '0'"
        yield f"for ii {Keywords.IN.value} 0 to {str(self.len_of_x_h_vector-1)} loop send_x_h_data(std_logic_vector(to_unsigned(ii, X_H_ADDR_WIDTH)), std_logic_vector(test_x_h_data(ii)), clock, x_config_en, x_config_addr, x_config_data)"
        yield f"wait for 10 ns"
        yield f"{Keywords.END.value} loop"
        yield f"for ii {Keywords.IN.value} 0 to {str(self.len_of_cell_vector-1)} loop send_c_data(std_logic_vector(to_unsigned(ii, HIDDEN_ADDR_WIDTH)), std_logic_vector(test_c_data(ii)), clock, c_config_en, c_config_addr, c_config_data)"
        yield f"wait for 10 ns"
        yield f"{Keywords.END.value} loop"
        yield f"enable <= '1'"
        yield f"wait until done = '1'"
        yield f"wait for 1*clk_period"
        yield f"enable <= '0'"
        yield f"-- reference h_out: {str(self.reference_h_out)}"
        yield f"for ii in 0 to {str(self.len_of_h_vector-1)} loop h_out_addr <= std_logic_vector(to_unsigned(ii, HIDDEN_ADDR_WIDTH))"
        yield f"h_out_en <= '1'"
        yield f"wait for 2*clk_period"
        yield f'report "The value of h_out(" & integer\'image(ii)& ") is " & integer\'image(to_integer(signed(h_out_data)))'
        yield f"{Keywords.END.value} loop"
        yield f"wait for 10*clk_period"

    def __call__(self) -> Iterable[Code]:
        yield from iter(self)



"""
The module contains classes and functions for generating vhdl code.
We provide code generators for the subset of vhdl that we need for implementing
our neural network accelerators and test benches. We stick closely to the vhdl
formal grammar with our class names.

The core of this module is the `CodeGenerator`. Code generators are callables that return `Code`.
`Code` is an iterable of strings. Depending on complexity we define syntactic components of the vhdl
grammar as `CodeGenerator`s. The class can then be used to set up and configure a function that yields lines
of code as strings.
"""
from enum import Enum
from itertools import chain, filterfalse
from typing import Callable, Iterable, Literal, Optional, Sequence, Union

Identifier = str
Code = Iterable[str]
CodeGenerator = Callable[[], Code]
CodeGeneratorCompatible = Code | CodeGenerator | str


class Keywords(Enum):
    IS = "is"
    END = "end"
    ENTITY = "entity"
    COMPONENT = "component"
    PORT = "port"
    GENERIC = "generic"
    IN = "in"
    OUT = "out"
    INOUT = "inout"
    BUFFER = "buffer"
    LINKAGE = "linkage"
    INTEGER = "integer"
    STD_LOGIC = "std_logic"
    SIGNAL = "signal"
    ARCHITECTURE = "architecture"
    OF = "of"
    SIGNED = "signed"
    BEGIN = "begin"
    PROCESS = "process"
    STD_LOGIC_VECTOR = "std_logic_vector"


class DataType(Enum):
    INTEGER = Keywords.INTEGER.value
    STD_LOGIC = Keywords.STD_LOGIC.value
    SIGNED = Keywords.SIGNED.value
    STD_LOGIC_VECTOR = Keywords.STD_LOGIC_VECTOR.value


class Mode(Enum):
    IN = Keywords.IN.value
    OUT = Keywords.OUT.value
    INOUT = Keywords.INOUT.value
    BUFFER = Keywords.BUFFER.value


class _DesignUnitForEntityAndComponent:
    def __init__(
        self, identifier: str, design_type: Literal[Keywords.ENTITY, Keywords.PORT]
    ):
        self.identifier = identifier
        self._generic_list = InterfaceList()
        self._port_list = InterfaceList()
        self.type = design_type

    @property
    def generic_list(self):
        return self._generic_list

    @generic_list.setter
    def generic_list(self, value):
        self._generic_list = InterfaceList(value)

    @property
    def port_list(self):
        return self._port_list

    @port_list.setter
    def port_list(self, value):
        self._port_list = InterfaceList(value)

    def _header(self) -> Code:
        if len(self.generic_list) > 0:
            yield from _clause(Keywords.GENERIC, self._generic_list())
        if len(self.port_list) > 0:
            yield from _clause(Keywords.PORT, self._port_list())

    def __call__(self) -> Code:
        yield f"{self.type.value} {self.identifier} {Keywords.IS.value}"
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.END.value} {self.type.value} {self.identifier};"


class Entity(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.ENTITY)


class ComponentDeclaration(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.COMPONENT)


class Architecture:
    def __init__(self, design_unit: str):
        self.design_unit = design_unit
        self._architecture_declaration_list = InterfaceList()
        self._architecture_component_list = InterfaceList()
        self._architecture_assignment_list = InterfaceList()
        self._architecture_process_list = InterfaceList()
        self._architecture_port_map_list = InterfaceList()
        self._architecture_assignment_at_end_of_declaration_list = InterfaceList()
        self._architecture_statement_part = InterfaceList()

    @property
    def architecture_declaration_list(self):
        return self._architecture_declaration_list

    @architecture_declaration_list.setter
    def architecture_declaration_list(self, value):
        self._architecture_declaration_list = InterfaceList(value)

    @property
    def architecture_statement_part(self):
        return self._architecture_statement_part

    @architecture_statement_part.setter
    def architecture_statement_part(self, value):
        self._architecture_statement_part = value

    @property
    def architecture_component_list(self):
        return self._architecture_component_list

    @architecture_component_list.setter
    def architecture_component_list(self, value):
        self._architecture_component_list = InterfaceList(value)

    @property
    def architecture_assignment_list(self):
        return self._architecture_assignment_list

    @architecture_assignment_list.setter
    def architecture_assignment_list(self, value):
        self._architecture_assignment_list = InterfaceList(value)

    @property
    def architecture_process_list(self):
        return self._architecture_process_list

    @architecture_process_list.setter
    def architecture_process_list(self, value):
        self._architecture_process_list = InterfaceList(value)

    @property
    def architecture_port_map_list(self):
        return self._architecture_port_map_list

    @architecture_port_map_list.setter
    def architecture_port_map_list(self, value):
        self._architecture_port_map_list = InterfaceList(value)

    @property
    def architecture_assignment_at_end_of_declaration_list(self):
        return self._architecture_assignment_at_end_of_declaration_list

    @architecture_assignment_at_end_of_declaration_list.setter
    def architecture_assignment_at_end_of_declaration_list(self, value):
        self._architecture_assignment_at_end_of_declaration_list = InterfaceList(value)

    def __call__(self) -> Code:
        yield f"{Keywords.ARCHITECTURE.value} rtl {Keywords.OF.value} {self.design_unit} {Keywords.IS.value}"
        if len(self._architecture_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self._architecture_declaration_list(), semicolon_last=True
                )
            )
        if len(self._architecture_component_list) > 0:
            yield from _filter_empty_lines(self._architecture_component_list())
        yield f"{Keywords.BEGIN.value}"
        if len(self._architecture_assignment_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self._architecture_assignment_list(), semicolon_last=True
                )
            )
        if len(self._architecture_process_list) > 0:
            yield from _filter_empty_lines(self.architecture_process_list())
        if len(self._architecture_port_map_list) > 0:
            yield from _filter_empty_lines(self.architecture_port_map_list())
        if len(self._architecture_assignment_at_end_of_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self._architecture_assignment_at_end_of_declaration_list(),
                    semicolon_last=True,
                )
            )

        if self._architecture_statement_part:
            yield from _filter_empty_lines(self._architecture_statement_part())
        yield f"{Keywords.END.value} {Keywords.ARCHITECTURE.value} rtl;"


class Process:
    def __init__(
        self,
        identifier: str,
        input_name: str = None,
        lookup_table_generator_function: CodeGenerator = None,
    ):
        self.identifier = identifier
        self.process_declaration_list = []
        self.process_statements_list = []
        self.lookup_table_generator_function = lookup_table_generator_function
        self.input = input_name

    def _header(self) -> Code:
        if len(self.process_declaration_list) > 0:
            yield from _append_semicolons_to_lines(self.process_declaration_list)

    def _body(self) -> Code:
        if len(self.process_statements_list) > 0:
            yield from _append_semicolons_to_lines(self.process_statements_list)
        if self.lookup_table_generator_function:
            yield from self.lookup_table_generator_function()

    def __call__(self) -> Code:
        if self.input:
            yield f"{self.identifier}_{Keywords.PROCESS.value}: {Keywords.PROCESS.value}({self.input})"
        else:
            yield f"{self.identifier}_{Keywords.PROCESS.value}: {Keywords.PROCESS.value}"
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.BEGIN.value}"
        yield from _filter_empty_lines(self._body())
        yield f"{Keywords.END.value} {Keywords.PROCESS.value} {self.identifier}_{Keywords.PROCESS.value};"


class ContextClause:
    def __init__(self, library_clause, use_clause):
        self._use_clause = use_clause
        self._library_clause = library_clause

    def __call__(self):
        yield from self._library_clause()
        yield from self._use_clause()


class UseClause:
    def __init__(self, selected_names: list[str]):
        self._selected_names = selected_names

    def __call__(self):
        def prefix_use(line: str):
            return f"use {line}"

        yield from _append_semicolons_to_lines(map(prefix_use, self._selected_names))


class LibraryClause:
    def __init__(self, logical_name_list: list[str]):
        self._logical_name_list = logical_name_list

    def __call__(self):
        yield from _append_semicolons_to_lines(
            ["library {}".format(", ".join(self._logical_name_list))]
        )


class InterfaceConstrained:
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        range: Optional[str | int],
        mode: Optional[Mode],
        value: Optional[str | int],
        declaration_type: Optional[str],
    ):
        self._identifier = identifier
        self._range = f"({range})" if range else ""
        self._identifier_type = identifier_type
        self._mode = f" {mode.value} " if mode else " "
        self._value = f" := {value}" if value else ""
        self._declaration_type = f"{declaration_type} " if declaration_type else ""

    def __call__(self) -> Code:
        yield from (
            f"{self._declaration_type}{self._identifier} :"
            f"{self._mode}{self._identifier_type.value}{self._range}{self._value}",
        )


class InterfaceSignal(InterfaceConstrained):
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        range: Optional[str | int] = None,
        mode: Optional[Mode] = None,
        value: Optional[str | int] = None,
    ):
        super().__init__(
            identifier, identifier_type, range, mode, value, declaration_type="signal"
        )


class InterfaceVariable(InterfaceConstrained):
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        range: Optional[str | int] = None,
        mode: Optional[Mode] = None,
        value: Optional[str | int] = None,
    ):
        super().__init__(
            identifier, identifier_type, range, mode, value, declaration_type=None
        )


class CodeGeneratorConcatenation(Sequence[CodeGenerator]):
    def __len__(self) -> int:
        return len(self.interface_generators)

    def __getitem__(self, item) -> CodeGenerator:
        return self.interface_generators.__getitem__(item)

    def __init__(self, *interfaces: CodeGeneratorCompatible):
        self.interface_generators = [
            _unify_code_generators(interface) for interface in interfaces
        ]

    def append(self, interface: CodeGeneratorCompatible) -> None:
        self.interface_generators.append(_unify_code_generators(interface))

    def __call__(self) -> Code:
        yield from chain.from_iterable(
            (interface() for interface in self.interface_generators)
        )


class PortMap:
    def __init__(self, map_name, component_name):
        self.map_name = map_name
        self.component_name = component_name
        self._signal_list = InterfaceList()
        self._generic_map_list = InterfaceList()

    @property
    def signal_list(self):
        return self._signal_list

    @signal_list.setter
    def signal_list(self, value):
        self._signal_list = InterfaceList(value)

    @property
    def generic_map_list(self):
        return self._generic_map_list

    @generic_map_list.setter
    def generic_map_list(self, value):
        self._generic_map_list = InterfaceList(value)

    def __call__(self) -> Code:
        yield f"{self.map_name}: {self.component_name}"
        if len(self._generic_map_list) > 0:
            yield f"generic map ("
            yield from _filter_empty_lines(_add_comma(self._generic_map_list()))
            yield ")"
        yield f"port map ("
        yield from _filter_empty_lines(_add_comma(self._signal_list()))
        yield ");"


class InterfaceList(CodeGeneratorConcatenation):
    pass


InterfaceDeclaration = Union[
    "InterfaceObjectDeclaration",
    "InterfaceTypeDeclaration",
    "InterfaceSubprogramDeclaration",
    "InterfacePackageDeclaration",
]

InterfaceObjectDeclaration = Union[
    "InterfaceConstantDeclaration",
    "InterfaceSignalDeclaration",
    "InterfaceVariableDeclaration",
    "InterfaceFileDeclaration",
]

ClauseType = Literal[Keywords.GENERIC, Keywords.PORT]


def indent(line: str) -> str:
    return "".join(["\t", line])


def _add_semicolons(lines: Code, semicolon_last: bool = False) -> Code:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp[:-1])
    yield f"{temp[-1]};" if semicolon_last else f"{temp[-1]}"


def _add_is(lines: Code) -> Code:
    yield from (f"{line} is" for line in lines)


def _add_comma(lines: Code, comma_last: bool = False) -> Code:
    temp = tuple(lines)
    yield from (f"{line}," for line in temp[:-1])
    yield f"{temp[-1]}," if comma_last else f"{temp[-1]}"


def _append_semicolons_to_lines(lines: Code) -> Code:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp)


def _clause(clause_type: ClauseType, interfaces: Code) -> Code:
    yield f"{clause_type.value} ("
    yield from _filter_empty_lines(_add_semicolons(interfaces))
    yield ");"


def _filter_empty_lines(lines: Code) -> Code:
    return filterfalse(_line_is_empty, lines)


def _line_is_empty(line: str) -> bool:
    return len(line) == 0


def _join_lines(lines) -> str:
    return "\n".join(lines)


def _empty_code_generator() -> Code:
    return []


# noinspection PyPep8Naming
def _wrap_in_IS_END_block(
    block_type: Keywords, block_identifier: Identifier, lines: Code
) -> Code:
    yield f"{block_type.value} {block_identifier} {Keywords.IS.value}"
    yield from _filter_empty_lines(lines)
    yield f"{Keywords.END.value} {block_type.value} {block_identifier};"


def _wrap_string_into_code_generator(string: str) -> CodeGenerator:
    def wrapped():
        return (string,)

    return wrapped


def _wrap_code_into_code_generator(code: Code) -> CodeGenerator:
    def wrapped():
        return code

    return wrapped


def _unify_code_generators(generator: CodeGeneratorCompatible) -> CodeGenerator:
    if isinstance(generator, str):
        return _wrap_string_into_code_generator(generator)
    elif isinstance(generator, Iterable):
        return _wrap_code_into_code_generator(generator)
    elif isinstance(generator, Callable):
        return generator
    else:
        raise ValueError


class Procedure:
    def __init__(self, identifier: str):
        self.identifier = identifier
        self._declaration_list = InterfaceList()
        self._declaration_list_with_is = InterfaceList()
        self._statement_list = InterfaceList()

    @property
    def declaration_list(self):
        return self._declaration_list

    @declaration_list.setter
    def declaration_list(self, value):
        self._declaration_list = InterfaceList(value)

    @property
    def declaration_list_with_is(self):
        return self._declaration_list_with_is

    @declaration_list_with_is.setter
    def declaration_list_with_is(self, value):
        self._declaration_list_with_is = InterfaceList(value)

    @property
    def statement_list(self):
        return self._statement_list

    @statement_list.setter
    def statement_list(self, value):
        self._statement_list = InterfaceList(value)

    def __call__(self):
        yield f"procedure {self.identifier} ("
        if len(self._declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(self._declaration_list(), semicolon_last=True)
            )
        if len(self._declaration_list_with_is) > 0:
            yield from _filter_empty_lines(_add_is(self._declaration_list_with_is()))
        yield f"{Keywords.BEGIN.value}"
        if len(self._statement_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(self._statement_list(), semicolon_last=True)
            )
        yield f"{Keywords.END.value} {self.identifier};"


def hex_representation(hex_value: str) -> str:
    return f'x"{hex_value}"'


def bin_representation(bin_value: str) -> str:
    return f'"{bin_value}"'



import json
from typing import Iterable

import torch

from elasticai.creator.precomputation.precomputation import (
    JSONEncoder,
    Precomputation,
    precomputable,
)
from elasticai.creator.tags_utils import Module, get_tags
from elasticai.creator.tests.tensor_test_case import TensorTestCase


class DummyModule:
    def __init__(self):
        self.call = lambda x: x

    @property
    def training(self) -> bool:
        return False

    # noinspection PyMethodMayBeStatic
    def extra_repr(self) -> str:
        return ""

    # noinspection PyMethodMayBeStatic
    def named_children(self) -> Iterable[tuple[str, "Module"]]:
        yield from ()

    def __call__(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        return self.call(x)


class PrecomputationTest(TensorTestCase):
    def test_precompute(self):
        module = DummyModule()

        def call(x: torch.Tensor) -> torch.Tensor:
            return x * 2

        module.call = call
        precompute = Precomputation(
            module=module, input_domain=torch.tensor([[1], [1]])
        )
        precompute()
        self.assertTensorEquals(torch.tensor([[2], [2]]), precompute.output)

    def test_precomputation_is_json_encodable(self):
        module = DummyModule()
        precompute = Precomputation(module=module, input_domain=torch.tensor([[]]))
        precompute()
        actual = json.dumps(precompute, cls=JSONEncoder)
        expected = """{"description": [], "shape": [0], "x": [[]], "y": [[]]}"""
        self.assertEqual(expected, actual)

    def test_precomputation_is_json_encodable_for_shape1x1(self):
        module = DummyModule()
        precompute = Precomputation(module=module, input_domain=torch.tensor([[[1]]]))
        precompute()
        actual = json.dumps(precompute, cls=JSONEncoder)
        expected = (
            """{"description": [], "shape": [1, 1], "x": [[[1]]], "y": [[[1]]]}"""
        )
        self.assertEqual(expected, actual)

    def test_precomputation_is_json_encoding_correctly_writes_in_and_output(self):
        module = DummyModule()

        def call(x: torch.Tensor) -> torch.Tensor:
            return torch.flatten(x * 3)

        module.call = call
        precompute = Precomputation(module=module, input_domain=torch.tensor([[[1]]]))
        precompute()
        actual = json.dumps(precompute, cls=JSONEncoder)
        expected = """{"description": [], "shape": [1, 1], "x": [[[1]]], "y": [3]}"""
        self.assertEqual(expected, actual)

    def test_precomputable_tag(self):
        input_generator = [-1, 1]

        @precomputable(input_shape=[-1, 1], input_generator=input_generator)
        class MyCustomModule(torch.nn.Module):
            def __init__(self):
                super().__init__()

        module = MyCustomModule()
        tags = get_tags(module)

        self.assertSequenceEqual(tags["precomputable"]["input_shape"], [-1, 1])



from unittest import TestCase

from torch import Tensor


class TensorTestCase(TestCase):
    def assertTensorEquals(self, expected: Tensor, actual: Tensor):
        equals = expected.equal(actual)
        if not equals:
            raise AssertionError(
                "Tensors differ: {expected} != {actual}".format(
                    expected=expected, actual=actual
                )
            )



import os
from argparse import ArgumentParser

from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point
from elasticai.creator.vhdl.rom import Rom
from elasticai.creator.vhdl.vhdl_formatter.vhdl_formatter import format_vhdl


def main():
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--path",
        help="path to folder for generated vhd files",
        required=True,
    )
    args = arg_parser.parse_args()
    if not os.path.isdir(args.path):
        os.mkdir(args.path)
    rom_name = "bi_rom"
    file_path = os.path.join(args.path, f"{rom_name}.vhd")

    data_width = 12
    frac_width = 4
    # biases for the input gate
    bi = [1.1]

    with open(file_path, "w") as writer:
        rom = Rom(
            rom_name=rom_name,
            values=float_values_to_fixed_point(
                bi, total_bits=data_width, frac_bits=frac_width
            ),
            resource_option="auto",
        )
        for line in rom():
            writer.write(line)
            if line[-1] != "\n":
                writer.write("\n")

    format_vhdl(file_path=file_path)


if __name__ == "__main__":
    main()
